
import mcol

from .compound import Compound
from .recipe import Recipe

from mlog import setup_logger
logger = setup_logger('HIPPO')

class Reaction:

	"""A :class:`.Reaction` is a simplified representation of a synthetic pathway to create a product :class:`.Compound`. Reactants (also :class:`.Compound` objects) as well as a reaction type are required.

	:class:`.Reaction` objects should not be created directly. Instead use :meth:`.HIPPO.register_reaction` or :meth:`.HIPPO.reactions`
	"""

	_table = 'reaction'

	def __init__(self,
		db,
		id: int,
		type: str,
		product: int,
		product_yield: float,
	):
		
		self._db = db
		self._id = id
		self._type = type
		self._product = product
		self._product_yield = product_yield
		
	### FACTORIES

	### PROPERTIES

	@property
	def id(self) -> int:
		"""Returns the reaction ID"""
		return self._id

	@property
	def type(self) -> str:
		"""Returns the reaction tyoe"""
		return self._type

	@property
	def product(self):
		"""Returns the reaction's product :class:`.Compound`"""
		if isinstance(self._product, int):
			self._product = self.db.get_compound(id=self._product)
		return self._product

	@property
	def product_yield(self) -> float:
		"""Returns the reaction's product yield (fraction)"""
		return self._product_yield

	@property
	def db(self):
		"""Returns a pointer to the parent database"""
		return self._db

	@property
	def reactants(self):
		"""Returns a :class:`.CompoundSet` of the reactants"""
		from .cset import CompoundSet
		return CompoundSet(self.db, indices=self.reactant_ids)

	@property
	def reaction_str(self) -> str:
		"""Returns a string representing the reaction"""
		s = ' + '.join([str(r) for r in self.reactants])
		s = f'{s} -> {str(self.product)}'
		return s

	@property
	def reactant_ids(self) -> set:
		"""Returns a set of reactant ID's"""
		return set(v for v in self.get_reactant_ids())

	@property
	def reactant_str_ids(self):
		return str(tuple(self.reactant_ids)).replace(',)',')')

	@property
	def product_id(self) -> int:
		"""Returns the product ID"""
		return self.product.id
	
	@property
	def product_smiles(self):
		return self.product.smiles

	@property
	def reactant_smiles(self):
		return [r.smiles for r in self.reactants]

	@property
	def product_mol(self):
		return self.product.mol

	@property
	def reactant_mols(self):
		return [r.mol for r in self.reactants]

	@property
	def price_estimate(self):
		return self.db.get_reaction_price_estimate(reaction=self)

	### METHODS

	def get_reactant_amount_pairs(self) -> list[Compound]:
		"""Returns pairs of reactants and their amounts"""

		compound_ids = self.db.select_where(query='reactant_compound, reactant_amount', table='reactant', key='reaction', value=self.id, multiple=True)

		if compound_ids:
			reactants = [(self.db.get_compound(id=id),amount) for id, amount in compound_ids]
		else:
			return []

		return reactants

	def get_reactant_ids(self) -> list[Compound]:

		compound_ids = self.db.select_where(query='reactant_compound', table='reactant', key='reaction', value=self.id, multiple=True)

		if compound_ids:
			return [id for id, in compound_ids]
		else:
			return []
	
	def get_recipes(self, 
		amount: float = 1, # in mg
		debug: bool = False,
		pick_cheapest: bool = False,
		permitted_reactions = None,
	):
		"""Get a :class:`.Recipe` describing how to make the product"""

		from .recipe import Recipe
		return Recipe.from_reaction(self, amount=amount, debug=debug, pick_cheapest=pick_cheapest, permitted_reactions=permitted_reactions)

	def summary(self, amount=1, draw=True):
		"""Print a summary of this reaction's information"""

		print(f'id={self.id}')
		print(f'type={self.type}')
		print(f'product={self.product}')
		print(f'product_yield={self.product_yield}')

		reactants = self.get_reactant_amount_pairs()
		print(f'reactants={reactants}')
		
		print(f'price_estimate={self.price_estimate}')

		# print(f'Ingredients for {amount} mg of product:')
		# ingredients = self.get_ingredients(amount=amount)
		# print(ingredients)

		if draw:
			self.draw()

		# return self.get_recipe(amount)

	def draw(self):
		"""Draw the molecules involved in this reaction"""

		from molparse.rdkit import draw_grid

		reactants = self.reactants

		product = self.product

		mols = [r.mol for r in reactants]
		mols.append(product.mol)

		labels = [f'+ {r}' if i>0 else f'{r}' for i,r in enumerate(reactants)]
		labels.append(f'-> {product}')

		drawing = draw_grid(mols, labels=labels, highlightAtomLists=None)
		display(drawing)

	def check_chemistry(self, debug=False):
		from .chem import check_chemistry
		return check_chemistry(self.type, self.reactants, self.product, debug=debug)

	def check_reactant_availability(self, supplier: None | str = None, debug: bool = False):

		if supplier is not None:
			raise NotImplementedError

		if debug:
			logger.var('reaction', self.id)
			logger.var('reactants', self.reactant_ids)
			logger.var('supplier', supplier)

		if supplier is None:

			triples = self.db.execute(f'''
				SELECT reactant_compound, quote_id, reaction_id FROM reactant 
				LEFT JOIN quote ON quote_compound = reactant_compound
				LEFT JOIN reaction ON reaction_product = reactant_compound 
				WHERE reactant_reaction = {self.id}
			''').fetchall()

		else:

			triples = self.db.execute(f'''
				SELECT reactant_compound, quote_id, reaction_id FROM reactant 
				LEFT JOIN quote ON quote_compound = reactant_compound
				LEFT JOIN reaction ON reaction_product = reactant_compound 
				WHERE reactant_reaction = {self.id}
				AND quote_supplier = "{supplier}"
			''').fetchall()

		for reactant_compound, quote_id, reaction_id in triples:
			if quote_id or reaction_id:
				continue
			if debug:
				logger.warning(f'No quote or reaction for reactant={reactant_compound}')
			return False

		return True

	def get_dict(self, smiles=True, mols=True):

		"""Returns a dictionary representing this Reaction"""

		serialisable_fields = ['id','type', 'product_id', 'reactant_ids']

		data = {}
		for key in serialisable_fields:
			data[key] = getattr(self, key)

		if smiles:
			data['product_smiles'] = self.product_smiles
			data['reactant_smiles'] = self.reactant_smiles

		if mols:
			data['product_mol'] = self.product_mol
			data['reactant_mols'] = self.reactant_mols

		return data

	### DUNDERS

	def __str__(self):
		return f'R{self.id}'

	def __repr__(self):
		return f'{mcol.bold}{mcol.underline}{self}: {self.reaction_str} via {self.type}{mcol.unbold}{mcol.ununderline}'

	def __eq__(self, other):

		match other:
			case int():
				return self.id == other

			case Reaction():
				
				if self.type != other.type:
					return False

				if self.product != other.product:
					return False

				if self.reactant_ids != other.reactant_ids:
					return False

				return True

			case _:
				raise NotImplementedError


	def __hash__(self):
		return self.id




import molparse as mp
import plotly.express as px
import pandas as pd
import mout
import plotly.graph_objects as go
from tqdm import tqdm

import logging
logger = logging.getLogger('HIPPO')

'''

	ALL GRAPHS DEFINED HERE SHOULD:

	* Have a HIPPO logo
	* Include the target name in the title

'''

import functools

# hippo_graph decorator
def hippo_graph(func):

	@functools.wraps(func)
	def wrapper(animal, *args, logo='top right', **kwargs):

		wrapper_kwargs = {}
		wrapper_keys = ['show', 'html', 'pdf', 'png']
		for key in wrapper_keys:
			wrapper_kwargs[key] = kwargs.pop(key,None)

		fig = func(animal, *args, **kwargs)

		if not isinstance(fig, go.Figure):
			return fig

		if wrapper_kwargs['show']:
				fig.show()

		if wrapper_kwargs['html']:
			file = wrapper_kwargs['html']
			if not file.endswith('.html'):
				file = f'{file}.html'
			mp.write(file, fig)

		if wrapper_kwargs['pdf']:
			file = wrapper_kwargs['pdf']
			if not file.endswith('.pdf'):
				file = f'{file}.pdf'
			mp.write(file, fig)

		if wrapper_kwargs['png']:
			file = wrapper_kwargs['png']
			if not file.endswith('.png'):
				file = f'{file}.png'
			mp.write(file, fig)

		if not fig.layout.images and logo:
			add_hippo_logo(fig, position=logo)

		return fig
	
	return wrapper

@hippo_graph
def plot_tag_statistics(animal, color='type', subtitle=None, log_y=False, compounds=True, poses=True):

	plot_data = []

	for tag in animal.tags.unique:

		if compounds:
			num_compounds = len(animal.compounds.get_by_tag(tag=tag))
			data = dict(tag=tag, number=num_compounds, type='compounds')
			plot_data.append(data)

		if poses:
			num_poses = len(animal.poses.get_by_tag(tag=tag))
			data = dict(tag=tag, number=num_poses, type='poses')
			plot_data.append(data)
		
	fig = px.bar(plot_data, x='tag', y='number', color=color, log_y=log_y)

	title = 'Tag Statistics'

	if subtitle:
		title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'
	else:
		title = f'<b>{animal.name}</b>: {title}'
	
	fig.update_layout(title=title,title_automargin=False, title_yref='container', barmode='group')

	fig.update_layout(xaxis_title='Tag', yaxis_title='#')

	return fig

@hippo_graph
def plot_interaction_histogram(animal, poses, feature_metadata, subtitle=None,):

	df = animal._fingerprint_df(poses)

	plot_data = []
	
	for key in df.columns:
		count = int(df[key].sum())

		if not count:
			continue
		
		data = dict(str=key,count=count)

		data['family'] = feature_metadata[key]['family']
		data['res_name'] = feature_metadata[key]['res_name']
		data['res_number'] = feature_metadata[key]['res_number']
		data['res_chain'] = feature_metadata[key]['res_chain']
		data['atom_numbers'] = feature_metadata[key]['atom_numbers']

		data['res_name_number_chain_str'] = f"{feature_metadata[key]['res_name']} {feature_metadata[key]['res_number']} {feature_metadata[key]['res_chain']}"
	
		plot_data.append(data)

	plot_df = pd.DataFrame(plot_data)
	plot_df.sort_values(['res_chain', 'res_number', 'family'], inplace=True)
	plot_df

	fig = px.bar(plot_df, x='res_name_number_chain_str', y='count', color='family', hover_data=plot_df.columns)

	title='Leveraged protein features'

	if subtitle:
		title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'
	else:
		title = f'<b>{animal.name}</b>: {title}'
	
	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_layout(xaxis_title='Residue')
	fig.update_layout(yaxis_title='#Interactions')

	return fig

@hippo_graph
def plot_interaction_punchcard(animal, poses=None, subtitle=None, opacity=1.0, group='pose_id', ignore_chains=False):

	import plotly

	plot_data = []

	categoryarray = {}


	if ignore_chains:
		x = 'chain_res_name_number_str'
	else:
		x = 'res_name_number_str'

	poses = poses or animal.poses
	
	logger.var('poses', poses)

	for pose in poses:

		fingerprint = pose.fingerprint

		if not fingerprint:
			continue

		# loop over each interaction in the pose
		for key, value in fingerprint.items():
			if not value:
				continue
			
			data = dict(str=key,count=value)

			f = animal.db.get_feature(id=key)

			data['id'] = f.id
			data['family'] = f.family
			data['residue_name'] = f.residue_name
			data['residue_number'] = f.residue_number
			data['chain_name'] = f.chain_name
			data['atom_names'] = f.atom_names
			
			data['pose_name'] = pose.name
			data['pose_id'] = str(pose)

			# data['chain_res_name_number_str'] = f.chain_res_name_number_str
			data[x] = getattr(f, x)

			if data[x] not in categoryarray:
				categoryarray[data[x]] = (data[x], f.chain_name, f.residue_number)

			plot_data.append(data)

	plot_df = pd.DataFrame(plot_data)

	if not len(plot_df):
		logger.error('Plotting data is empty. Are the poses fingerprinted?')
		return None

	plot_df = plot_df.sort_values(group)

	title = 'Interaction Punch-Card'

	if subtitle:
		title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'
	else:
		title = f'<b>{animal.name}</b>: {title}'

	fig = px.scatter(plot_df, x=x, y='family',marginal_x='histogram',marginal_y='histogram', hover_data=plot_df.columns, color=group, title=title)
	
	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_layout(xaxis_title='Residue', yaxis_title='Feature Family')

	categoryarray = sorted([v for v in categoryarray.values()], key=lambda x: (x[1], x[2]))
	categoryarray = [v[0] for v in categoryarray]

	fig.update_xaxes(categoryorder='array', categoryarray=categoryarray)

	for trace in fig.data:
		if type(trace) == plotly.graph_objs._histogram.Histogram:
			trace.opacity = 1
			trace.xbins.size = 1
		else:
			trace['marker']['size'] = 10
			trace['marker']['opacity'] = opacity

	fig.update_layout(barmode='stack')
	fig.update_layout(scattermode='group', scattergap=0.75)

	# return add_hippo_logo(fig, in_plot=False)
	return add_punchcard_logo(fig)

@hippo_graph
def plot_residue_interactions(animal, poses, residue_number, subtitle=None, chain=None):

	assert not chain

	logger.debug(poses)
	logger.debug(residue_number)

	plot_data = []

	categoryarray = {}

	for pose in poses:

		fingerprint = pose.fingerprint

		if not fingerprint:
			continue

		# loop over each interaction in the pose
		for key, value in fingerprint.items():
			if not value:
				continue

			f = animal.db.get_feature(id=key)

			if chain and f.chain_name != chain:
				continue

			if residue_number != f.residue_number:
				continue

			data = dict(str=key,count=value)

			data['id'] = f.id
			data['family'] = f.family
			data['residue_name'] = f.residue_name
			data['residue_number'] = f.residue_number
			data['chain_name'] = f.chain_name
			data['atom_names'] = f.atom_names
			
			data['pose_name'] = pose.name
			data['pose_id'] = str(pose)
			
			data['residue_name_number'] = f'{f.residue_name} {f.residue_number}'

			if data['pose_name'] not in categoryarray:
				categoryarray[data['pose_name']] = [data['pose_name'], data['count']]
			else:
				categoryarray[data['pose_name']][1] += data['count']

			for _ in range(value):
				plot_data.append(data)

	fig = px.histogram(plot_data, x='pose_name', color='family')
	
	categoryarray = sorted([v for v in categoryarray.values()], key=lambda x: (-x[1], x[0]))
	categoryarray = [v[0] for v in categoryarray]

	fig.update_xaxes(categoryorder='array', categoryarray=categoryarray)

	if not subtitle:
		subtitle = f'#Poses={len(poses)}'

	title = f"Interactions w/ {data['residue_name_number']}"

	title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_xaxes(title='Pose')
	fig.update_yaxes(title='#Interactions')

	return fig


@hippo_graph
# def plot_building_blocks(animal, subtitle=None, cset='elabs', color='name_is_smiles'):
def plot_reactant_amounts(animal, subtitle=None, color='has_price_picker', named_only=False, most_common=None):

	# cset = animal.compound_sets[cset]

	# bbs = cset.get_building_blocks()

	bbs = animal.building_blocks

	mout.debug('making plot_data')
	plot_data = []
	for bb in bbs:
		d = bb.dict
		# if most_common and d['amount'] is not None:
			
		if not named_only or not d['name_is_smiles']:
			plot_data.append(d)

	if most_common:
		mout.debug('sorting')
		plot_data = sorted(plot_data, key=lambda x: x['amount'], reverse=True)[:most_used_number]

	fig = px.bar(plot_data, x='name', y='amount', color=color, hover_data=plot_data[0].keys())
	# fig = px.bar(plot_data, x='smiles', y='amount', color=color)

	title = 'Building Blocks'

	if not subtitle:
		# subtitle = f'"{cset.name}": #BBs={len(bbs)}'
		subtitle = f'#BBs={len(bbs)}'

	title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_layout(xaxis_title='Reactant', yaxis_title='#Reactions')

	return fig

@hippo_graph
# def plot_building_blocks(animal, subtitle=None, cset='elabs', color='name_is_smiles'):
def plot_reactant_price(animal, subtitle=None, amount=20):

	# cset = animal.compound_sets[cset]

	# bbs = cset.get_building_blocks()

	bbs = animal.building_blocks

	plot_data = []
	for bb in bbs:
		d = bb.dict
		if not d['has_price_picker']:
			continue

		d[f'price_{amount}mg'] = bb.get_price(amount)
		d[f'min_amount'] = bb.price_picker.min_amount

		plot_data.append(d)

	# fig = px.bar(plot_data, x='name', y=f'price_{amount}mg', color='lead_time', log_y=True, hover_data=plot_data[0].keys())
	fig = px.histogram(plot_data, x=f'price_{amount}mg', color='lead_time', hover_data=plot_data[0].keys())
	# fig = px.bar(plot_data, x='smiles', y='amount', color=color)

	title = 'Reactant Pricing'

	if not subtitle:
		# subtitle = f'"{cset.name}": #BBs={len(bbs)}'
		subtitle = f'#BBs={len(bbs)}'

	title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_layout(yaxis_title='Number of reactants', xaxis_title=f'Price for {amount}mg [$USD]')

	return fig

@hippo_graph
# def plot_building_blocks(animal, subtitle=None, cset='elabs', color='name_is_smiles'):
def plot_reactants_2d(animal, subtitle=None, amount=20):

	# cset = animal.compound_sets[cset]

	# bbs = cset.get_building_blocks()

	bbs = animal.building_blocks

	plot_data = []
	for bb in bbs:
		d = bb.dict
		if not d['has_price_picker']:
			continue

		d[f'price_{amount}mg'] = bb.get_price(amount)
		d[f'min_amount'] = bb.price_picker.min_amount

		plot_data.append(d)

	fig = px.scatter(plot_data, y='amount', x=f'price_{amount}mg', color='name', hover_data=plot_data[0].keys())
	# fig = px.bar(plot_data, x='smiles', y='amount', color=color)

	title = 'Building Blocks'

	if not subtitle:
		# subtitle = f'"{cset.name}": #BBs={len(bbs)}'
		subtitle = f'#BBs={len(bbs)}'

	title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_layout(scattermode='group', scattergap=0.75)

	fig.update_layout(yaxis_title='Quantity [mg]', xaxis_title=f'Price for {amount}mg [$USD]')

	return fig

@hippo_graph
# def plot_building_blocks(animal, subtitle=None, cset='elabs', color='name_is_smiles'):
def plot_building_blocks(animal, subtitle=None, color='name_is_smiles'):

	# cset = animal.compound_sets[cset]

	# bbs = cset.get_building_blocks()

	bbs = animal.building_blocks

	plot_data = []
	for bb in bbs:
		plot_data.append(bb.dict)

	fig = px.scatter(plot_data, x='name', y='max', color='amount')
	# fig = px.bar(plot_data, x='smiles', y='amount', color=color)

	title = 'Building Blocks'

	if not subtitle:
		# subtitle = f'"{cset.name}": #BBs={len(bbs)}'
		subtitle = f'#BBs={len(bbs)}'

	title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_layout(xaxis_title='Reactant', yaxis_title='Quantity')

	return fig

@hippo_graph
def plot_synthetic_routes(animal, subtitle=None, cset='elabs', color='num_reactants'):

	cset = animal.compound_sets[cset]

	plot_data = []
	for reax in cset.reactions:
		plot_data.append(reax.dict)

	# fig = px.bar(plot_data, x='name', y='amount', color=color)
	fig = px.histogram(plot_data, x='type', color=color)

	title = 'Synthetic Routes'

	if not subtitle:
		subtitle = f'"{cset.name}": #compounds={len(cset)}'

	title = f'<b>{animal.name}</b>: {title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container')

	fig.update_layout(xaxis_title='Compound', yaxis_title='#Routes')

	return fig

@hippo_graph
def plot_numbers(animal, subtitle=None):

	'''
		- y-axis: numbers
		- x-categories
			* hits
			* hit poses
			* bases
			* base poses
			* elabs
			* elab poses
			* BBs (total)
			* BBs (in enamine)
	'''

	# cset = animal.compound_sets[cset]

	plot_data = [
		dict(category='Experimental Hits', number=len(animal.hits), type='compound'),
		dict(category='Experimental Hits', number=len(animal.hits.poses), type='poses'),
		dict(category='Base compounds', number=len(animal.bases), type='compound'),
		dict(category='Base compounds', number=len(animal.bases.poses), type='poses'),
		dict(category='Syndirella Elaborations', number=len(animal.elabs), type='compound'),
		dict(category='Syndirella Elaborations', number=len(animal.elabs.poses), type='poses'),
		dict(category='Unique Reactants', number=len(animal.building_blocks), type='compound'),
	]

	fig = px.bar(plot_data, x='category', y='number', log_y=True, color='type')

	title = 'Compounds & Poses'

	title = f'<b>{animal.name}</b>: {title}<br>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container', barmode='group')

	fig.update_layout(xaxis_title=None, yaxis_title='Log(Quantity)')

	return fig

@hippo_graph
def plot_compound_property(animal, prop, compounds=None, style='bar', null=None):

	"""
	Get an arbitrary property from all the compounds in animal.compounds
	
	If one property, plot a 1D histogram
	If 2D plot a bar/scatter

	"""

	if not isinstance(prop, list):
		prop = [prop]

	plot_data = []

	if not compounds:
		compounds = animal.compounds

	if len(compounds) > 1000:
		compounds = tqdm(compounds)

	for comp in compounds:

		data = comp.dict

		for p in prop:

			# has attr

			if p not in data:

				# get attr
				if hasattr(comp,p):
					v = getattr(comp,p)
					
				elif p in (m := comp.metadata):
				    v = m[p]
					
				else:
					v = null

				data[p] = v

		plot_data.append(data)

	if len(prop) == 1:

		title = f'Compound {prop[0]}'

		fig = px.histogram(plot_data, x=prop[0])
		
		fig.update_layout(xaxis_title=prop[0], yaxis_title='Quantity')

	elif len(prop) == 2:

		hover_data = prop + ['smiles']

		title = f'Compound {prop[0]} vs {prop[1]}'
		
		func = eval(f'px.{style}')
		fig = func(plot_data, x=prop[0], y=prop[1], hover_data=hover_data)

		fig.update_layout(xaxis_title=prop[0], yaxis_title=prop[1])

	else:
		mout.error('Unsupported', code='plotting.plot_compound_property.1')

	title = f'<b>{animal.name}</b>: {title}<br>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container', barmode='group')

	return fig

@hippo_graph
def plot_pose_property(animal, prop, poses=None, style='scatter', title=None, null=None, color=None, log_y=False, subtitle=None, data_only=False, **kwargs):

	"""
	Get an arbitrary property from all the poses in animal.poses
	
	If one property, plot a 1D histogram
	If 2D plot a scatter plot

	"""

	# fetch these directly from the database
	if prop in ['energy_score', 'distance_score']:

		p = prop
		
		if not poses:
			# great all poses!
			poses = animal.poses
			n_poses = len(poses)
			logger.out(f'Querying database for {n_poses} poses...')
			field = f'pose_{p}'
			title = title or f'{p} of all poses'
			query = animal.db.select_where(table='pose', query=field, key=f'{field} is not NULL', multiple=True)
			
		else:

			# subset of poses
			assert poses.table == 'pose', f'{poses=} is not a set of Pose objects'
			n_poses = len(poses)
			logger.out(f'Querying database for {n_poses} poses...')
			field = f'pose_{p}'
			title = title or f'{p} of pose subset'
			query = animal.db.select_where(table='pose', query=field, key=f'{field} is not NULL and pose_id in {poses.str_ids}', multiple=True)
			
		plot_data = [{p:v} for v, in query]

		if p == 'energy_score':
			subtitle = subtitle or f'#poses={n_poses}, energy_score < 0 = {len([None for d in plot_data if d[p] < 0])/n_poses:.1%}'
		elif p == 'distance_score':
			subtitle = subtitle or f'#poses={n_poses}, distance_score < 2 = {len([None for d in plot_data if d[p] < 2])/n_poses:.1%}'
		else:
			subtitle = subtitle or f'#poses={n_poses}'

		prop = [prop]

	elif prop == ['energy_score', 'distance_score'] or prop == ['distance_score', 'energy_score']:
		
		query = f'pose_id, pose_distance_score, pose_energy_score'

		# hardcoded errorbars
		distance_score_err = 0.03
		energy_score_err = 6

		if color:
			query += f', {color}'
		
		if not poses:
			# great all poses!
			poses = animal.poses
			n_poses = len(poses)
			logger.out(f'Querying database for {n_poses} poses...')
			title = f'distance & energy scores of all poses'
			query = animal.db.select(table='pose', query=query, multiple=True)
			
		else:

			# subset of poses
			n_poses = len(poses)
			logger.out(f'Querying database for {n_poses} poses...')
			title = f'distance & energy scores of pose subset'
			query = animal.db.select_where(table='pose', query=query, key=f'pose_id in {poses.str_ids}', multiple=True)
			
		plot_data = []
		for q in query:
			d = {'id':q[0], 'distance_score':q[1], 'energy_score':q[2], 'distance_score_err':distance_score_err, 'energy_score_err':energy_score_err} 
			if color:
				d[color] = q[-1]
				if color == 'pose_compound':
					d[color] = f'C{d[color]}'

			plot_data.append(d)

		kwargs['error_x'] = 'energy_score_err'
		kwargs['error_y'] = 'distance_score_err'
		
		subtitle = subtitle or f'#poses={n_poses}'

	# elif prop == ['num_atoms_added', 'energy_score'] or prop == ['num_atoms_added', 'energy_score']:
	# 	logger.error('Use animal.plot_pose_risk_vs_placement')
	# 	raise NotImplementedError
		
	# 	query = f'pose_id, , pose_distance_score, pose_energy_score'
		
	# 	if not poses:
	# 		# great all poses!
	# 		poses = animal.poses
	# 		n_poses = len(poses)
	# 		logger.out(f'Querying database for {n_poses} poses...')
	# 		title = f'distance & energy scores of all poses'
	# 		query = animal.db.select(table='pose', query=query, multiple=True)
			
	# 	else:

	# 		# subset of poses
	# 		n_poses = len(poses)
	# 		logger.out(f'Querying database for {n_poses} poses...')
	# 		title = f'distance & energy scores of pose subset'
	# 		query = animal.db.select_where(table='pose', query=query, key=f'pose_id in {poses.str_ids}', multiple=True)
			
	# 	plot_data = [{'id':id, 'distance_score':v1, 'energy_score':v2 } for id,v1,v2 in query]
		
	# 	subtitle = f'#poses={n_poses}'
	
	else:

		if not poses:
			poses = animal.poses

		if prop == 'tags':

			plot_data = []
	
			for tag in poses.tags:
	
				num_poses = len(poses.get_by_tag(tag=tag))
				data = dict(tag=tag, number=num_poses)
				plot_data.append(data)
				
			fig = px.bar(plot_data, x='tag', y='number', color=color, log_y=log_y)
	
			title = 'Tag Statistics'
			
			fig.update_layout(title=title,title_automargin=False, title_yref='container', barmode='group')
	
			fig.update_layout(xaxis_title='Tag', yaxis_title='#')
	
			return fig

		if not isinstance(prop, list):
			prop = [prop]
	
		plot_data = []
		if len(poses) > 1000:
			poses = tqdm(poses)
	
		for pose in poses:
			if len(prop) > 1:
				data = dict(id=pose.id)
			else:
				data = {}
	
			for p in prop:
				if p not in data:
					# get attr
					if hasattr(pose,p):
						v = getattr(pose,p)
		
					elif p in (m := pose.metadata):				
						v = m[p]
					
					else:
						v = null
	
					data[p] = v

			if color:
				if color not in data:
					# get attr
					if hasattr(pose,color):
						v = getattr(pose,color)
		
					elif color in (m := pose.metadata):				
						v = m[color]
					
					else:
						v = null
	
					data[color] = v
	
			plot_data.append(data)

	if data_only:
		return plot_data
	
	hover_data = ['id'] #, 'alias', 'inchikey'] #, 'tags', 'inspirations']

	if len(prop) == 1:

		title = title or f'Pose {prop[0]}'

		fig = px.histogram(plot_data, x=prop[0], hover_data=None, color=color, **kwargs)
		
		fig.update_layout(xaxis_title=prop[0], yaxis_title='Quantity')

	elif len(prop) == 2:
		
		if style == 'histogram':

			x = [d[prop[0]] for d in plot_data]
			y = [d[prop[1]] for d in plot_data]

			fig = go.Figure(go.Histogram2d(x=x, y=y, **kwargs))

		else:
			
			if style == 'bar':
				style='scatter'
			
			func = eval(f'px.{style}')
			fig = func(plot_data, x=prop[0], y=prop[1], color=color, hover_data=hover_data, **kwargs)

		title = title or f'Pose {prop[0]} vs {prop[1]}'
		fig.update_layout(xaxis_title=prop[0], yaxis_title=prop[1])

	else:
		mout.error('Unsupported', code='plotting.plot_pose_property.1')

	title = title or f'<b>{animal.name}</b>: {title}<br>'
	
	if subtitle:
		title = f'{title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title,title_automargin=False, title_yref='container', barmode='group')

	return fig

@hippo_graph
def plot_compound_availability(animal, compounds=None, title=None, subtitle=None):

	from .cset import CompoundTable
	
	compounds = compounds or animal.compounds

	match compounds:
		case CompoundTable():
			pairs = animal.db.select(table='quote', query='DISTINCT quote_supplier, quote_catalogue', multiple=True)
		
			plot_data = []
			for supplier, catalogue in pairs:
		
				if catalogue is None:
					catalogue = 'None'
					cat_str = 'NULL'
				else:
					cat_str = f'"{catalogue}"'

				count, = animal.db.select_where(table='quote', query='COUNT(DISTINCT quote_compound)', key=f'quote_supplier IS "{supplier}" AND quote_catalogue IS {cat_str}')

				plot_data.append(dict(supplier=supplier, catalogue=catalogue, count=count))
				
		case _:
			raise NotImplementedError
		
	fig = px.bar(plot_data, x='catalogue', y='count', color='supplier')

	title = 'Compound availability'
	
	title = title or f'<b>{animal.name}</b>: Compound availability<br>'
	
	if subtitle:
		title = f'{title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title) #,title_automargin=False, title_yref='container', barmode='group')

	return fig

@hippo_graph
def plot_compound_price(animal, compounds=None, min_amount=1, subtitle=None, title=None, style='histogram', **kwargs):

	from .cset import CompoundTable
	import numpy as np
	
	compounds = compounds or animal.compounds

	match compounds:
		case CompoundTable():

			if style=='scatter':

				sql = '''
				SELECT quote_compound, quote_amount, MIN(quote_price), quote_lead_time, compound_smiles, COUNT(DISTINCT reactant_reaction)
				FROM quote 
				INNER JOIN compound ON quote.quote_compound = compound.compound_id
				INNER JOIN reactant ON quote.quote_compound = reactant.reactant_compound
				WHERE quote_amount >= {min_amount}
				GROUP BY quote_compound
				'''.format(min_amount=min_amount)

				results = animal.db.execute(sql).fetchall()

				n_compounds = len(results)

				plot_data = []
				for compound_id, amount, price, lead_time, smiles, num_reactions in results:
					plot_data.append(dict(compound_id=compound_id, min_price=price, quoted_amount=amount, lead_time=lead_time, smiles=smiles, num_reactions=num_reactions, log_price_per_reaction=np.log(price/num_reactions), price_per_reaction=price/num_reactions))

			else:
				data = animal.db.select_where(table='quote', query='quote_amount, MIN(quote_price)', key=f'quote_amount >= {min_amount} GROUP BY quote_compound', multiple=True)

				n_compounds = len(data)

				plot_data = []
				for amount, price in data:
					plot_data.append(dict(min_price=price, quoted_amount=amount))
	
		case _:
			raise NotImplementedError('CompoundSet not yet supported')

	plot_data = sorted(plot_data, key=lambda x: x['quoted_amount'])

	match style:
		
		case 'histogram':
			fig = px.histogram(plot_data, color='quoted_amount', x='min_price', **kwargs)
		
		case 'violin':
			fig = px.violin(plot_data, color='quoted_amount', x='min_price', **kwargs)
		
		case 'scatter':
			fig = px.scatter(plot_data, color='log_price_per_reaction', x='min_price', y='lead_time', hover_data=plot_data[0].keys(), **kwargs)
		
		case _:
			raise NotImplementedError(f'{style=}')

	subtitle = subtitle or f'#compounds={n_compounds}, {min_amount=} mg'
	
	title = title or f'<b>{animal.name}</b>: Compound price<br>'
	
	if subtitle:
		title = f'{title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title) #,title_automargin=False, title_yref='container', barmode='group')

	return fig

@hippo_graph
def plot_reaction_funnel(animal, title=None, subtitle=None):

	compounds = animal.compounds

	data = dict(
		number=[compounds.num_reactants, compounds.num_intermediates, compounds.num_products],
		category=["Reactants", "Intermediates", "Products"]
	)
	
	fig = px.funnel(data, x='category', y='number')

	title = title or f'<b>{animal.name}</b>: Reaction statistics'

	if subtitle:
		title = f'{title}<br><sup><i>{subtitle}</i></sup>'

	fig.update_layout(title=title, title_automargin=False, title_yref='container')
	
	return fig

HIPPO_LOGO_URL = 'https://raw.githubusercontent.com/mwinokan/HIPPO/main/logos/hippo_logo_tightcrop.png'
HIPPO_HEAD_URL = 'https://raw.githubusercontent.com/mwinokan/HIPPO/main/logos/hippo_assets-02.png'

def add_hippo_logo(fig, in_plot=True, position='top right'):

	assert fig.layout.title.text, 'Figure must have a title to add the HIPPO logo'

	if in_plot:

		sizex=0.3
		sizey=0.3

		if 'top' in position:
			yanchor="top"
			y = 0.95
		elif 'bottom' in position:
			yanchor="bottom"
			y = 0.05
		else:
			yanchor="middle"
			y = 0.50

		if 'left' in position:
			xanchor = "left"
			x = 0.05
		elif 'right' in position:
			xanchor = "right"
			x = 0.95
		else:
			xanchor = "center"
			x = 0.50

		fig.add_layout_image(dict(
			source=HIPPO_LOGO_URL,
			xref="paper", yref="paper",
			# layer='below',
			x=x, y=y,
			sizex=sizex, sizey=sizey,
			xanchor=xanchor, yanchor=yanchor,
		))

		return fig

	has_legend = fig.layout.legend.title.text is not None
	fig.layout.margin.t = None

	if has_legend:

		fig.add_layout_image(dict(
			source="",
			xref="paper", yref="paper",
			
			x=1, y=1.05,
			sizex=0.4, sizey=0.4,
			xanchor="left", yanchor="bottom",
		))

	else:

		fig.add_layout_image(dict(
			source=HIPPO_LOGO_URL,
			xref="paper", yref="paper",
			
			x=1, y=1.05,
			sizex=0.3, sizey=0.3,
			xanchor="right", yanchor="bottom",
		))
	
	return fig

def add_punchcard_logo(fig):

	fig.add_layout_image(dict(
		source=HIPPO_HEAD_URL,
		xref="paper", yref="paper",
		
		x=1, y=1,
		sizex=0.25, sizey=0.25,
		xanchor="right", yanchor="top",
	))

	return fig



import mout
import mcol
import sqlite3
from sqlite3 import Error
from pprint import pprint
import json

from .compound import Compound
from .pose import Pose
from .reaction import Reaction
from .quote import Quote
from .metadata import MetaData
from .target import Target
from .feature import Feature

from .tools import inchikey_from_smiles

from pathlib import Path

import logging
logger = logging.getLogger('HIPPO')

class Database:

	"""Wrapper to connect to the HIPPO sqlite database."""

	def __init__(self, path: Path, animal) -> None:

		assert isinstance(path, Path)
		
		logger.debug('hippo3.Database.__init__()')

		self._path = path
		self._connection = None
		self._cursor = None
		self._animal = animal

		logger.debug(f'Database.path = {self.path}')

		try:
			path = path.resolve(strict=True)

		except FileNotFoundError:
			# create a blank database
			self.connect()
			self.create_blank_db()

		else:
			# existing database
			self.connect()
		
	### PROPERTIES

	@property
	def path(self) -> Path:
		"""Returns the path to the database file"""
		return self._path

	@property
	def connection(self):
		"""Returns a sqlite3.connection"""
		if not self._connection:
			self.connect()
		return self._connection
	
	@property
	def cursor(self):
		"""Returns a sqlite3.cursor"""
		if not self._cursor:
			self._cursor = self.connection.cursor()
		return self._cursor

	@property
	def total_changes(self):
		"""Return the total number of database rows that have been modified, inserted, or deleted since the database connection was opened."""
		return self.connection.total_changes

	### PUBLIC METHODS / API CALLS

	def close(self) -> None:
		"""Close the connection"""
		logger.debug('hippo3.Database.close()')
		if self.connection:
			self.connection.close()
		mout.success(f'Closed connection to {mcol.file}{self.path}')

	### GENERAL SQL

	def connect(self):
		"""Connect to the database"""
		logger.debug('hippo3.Database.connect()')

		conn = None

		try:
			conn = sqlite3.connect(self.path)

			logger.debug(f'{sqlite3.version=}')

			conn.enable_load_extension(True)
			conn.load_extension('chemicalite')
			conn.enable_load_extension(False)

			logger.success(f'Database connected @ {mcol.file}{self.path}')
		
		except sqlite3.OperationalError as e:

			if 'cannot open shared object file' in str(e):
				logger.error('chemicalite package not installed correctly')
			else:
				logger.exception(e)
			raise

		except Error as e:
			logger.exception(e)
			raise
		
		self._connection = conn
		self._cursor = conn.cursor()

	def execute(self, sql, payload=None):
		"""Execute arbitrary SQL"""
		try:
			if payload:
				return self.cursor.execute(sql, payload)
			else:
				return self.cursor.execute(sql)
		except Error as e:
			# logger.exception(e)
			raise

	def commit(self):
		"""Commit the changes"""
		# logger.debug('db.commit...')
		self.connection.commit()
		# raise Exception

	### CREATE TABLES

	def create_blank_db(self):
		"""Create a blank database"""

		logger.out('Creating blank database...')
		self.create_table_compound()
		self.create_table_inspiration()
		self.create_table_reaction()
		self.create_table_reactant()
		self.create_table_pose()
		self.create_table_tag()
		self.create_table_quote()
		self.create_table_target()
		self.create_table_pattern_bfp()
		self.create_table_feature()
		# self.create_table_morgan_bfp()

	def create_table_compound(self):
		"""Create the compound table"""
		logger.debug('HIPPO.Database.create_table_compound()')

		sql = """CREATE TABLE compound(
			compound_id INTEGER PRIMARY KEY,
			compound_inchikey TEXT,
			compound_alias TEXT,
			compound_smiles TEXT,
			compound_base INTEGER,
			compound_mol MOL,
			compound_pattern_bfp bits(2048),
			compound_morgan_bfp bits(2048),
			compound_metadata TEXT,
			FOREIGN KEY (compound_base) REFERENCES compound(compound_id),
			CONSTRAINT UC_compound_inchikey UNIQUE (compound_inchikey)
			CONSTRAINT UC_compound_alias UNIQUE (compound_alias)
			CONSTRAINT UC_compound_smiles UNIQUE (compound_smiles)
		);
		"""
			# CONSTRAINT UC_compound_morgan_bfp UNIQUE (compound_morgan_bfp)
			# CONSTRAINT UC_compound_pattern_bfp UNIQUE (compound_pattern_bfp)

		self.execute(sql)

	def create_table_inspiration(self):
		"""Create the inspiration table"""
		logger.debug('HIPPO.Database.create_table_inspiration()')

		sql = """CREATE TABLE inspiration(
			inspiration_original INTEGER,
			inspiration_derivative INTEGER,
			FOREIGN KEY (inspiration_original) REFERENCES pose(pose_id),
			FOREIGN KEY (inspiration_derivative) REFERENCES pose(pose_id),
			CONSTRAINT UC_inspiration UNIQUE (inspiration_original, inspiration_derivative)
		);
		"""

		self.execute(sql)

	def create_table_reaction(self):
		"""Create the reaction table"""
		logger.debug('HIPPO.Database.create_table_reaction()')

		sql = """CREATE TABLE reaction(
			reaction_id INTEGER PRIMARY KEY,
			reaction_type TEXT,
			reaction_product INTEGER,
			reaction_product_yield REAL,
			FOREIGN KEY (reaction_product) REFERENCES compound(compound_id)
		);
		"""

		self.execute(sql)

	def create_table_reactant(self):
		"""Create the reactant table"""
		logger.debug('HIPPO.Database.create_table_reactant()')

		sql = """CREATE TABLE reactant(
			reactant_amount REAL,
			reactant_reaction INTEGER,
			reactant_compound INTEGER,
			FOREIGN KEY (reactant_reaction) REFERENCES reaction(reaction_id),
			FOREIGN KEY (reactant_compound) REFERENCES compound(compound_id),
			CONSTRAINT UC_reactant UNIQUE (reactant_reaction, reactant_compound)
		);
		"""

		self.execute(sql)

	def create_table_pose(self):
		"""Create the pose table"""
		logger.debug('HIPPO.Database.create_table_pose()')

		sql = """CREATE TABLE pose(
			pose_id INTEGER PRIMARY KEY,
			pose_inchikey TEXT,
			pose_alias TEXT,
			-- pose_longname TEXT,
			pose_smiles TEXT,
			pose_reference INTEGER,
			pose_path TEXT,
			pose_compound INTEGER,
			pose_target INTEGER,
			pose_mol BLOB,
			pose_fingerprint BLOB,
			pose_energy_score REAL,
			pose_distance_score REAL,
			pose_metadata TEXT,
			FOREIGN KEY (pose_compound) REFERENCES compound(compound_id),
			CONSTRAINT UC_pose_alias UNIQUE (pose_alias)
			CONSTRAINT UC_pose_path UNIQUE (pose_path)
		);
		"""

		### snippet to convert python metadata dictionary with JSON
		# json.dumps(variables).encode('utf-8')
		# json.loads(s.decode('utf-8'))

		self.execute(sql)

	def create_table_tag(self):
		"""Create the tag table"""
		logger.debug('HIPPO.Database.create_table_tag()')

		sql = """CREATE TABLE tag(
			tag_name TEXT,
			tag_compound INTEGER,
			tag_pose INTEGER,
			FOREIGN KEY (tag_compound) REFERENCES compound(compound_id),
			FOREIGN KEY (tag_pose) REFERENCES pose(pose_id),
			CONSTRAINT UC_tag_compound UNIQUE (tag_name, tag_compound)
			CONSTRAINT UC_tag_pose UNIQUE (tag_name, tag_pose)
		);
		"""

		self.execute(sql)

	def create_table_quote(self):
		"""Create the quote table"""
		logger.debug('HIPPO.Database.create_table_quote()')

		sql = """CREATE TABLE quote(
			quote_id INTEGER PRIMARY KEY,
			quote_smiles TEXT,
			quote_amount REAL,
			quote_supplier TEXT,
			quote_catalogue TEXT,
			quote_entry TEXT,
			quote_lead_time INTEGER,
			quote_price REAL,
			quote_currency TEXT,
			quote_purity REAL,
			quote_date TEXT,
			quote_compound INTEGER,
			FOREIGN KEY (quote_compound) REFERENCES compound(compound_id),
			CONSTRAINT UC_quote UNIQUE (quote_amount, quote_supplier, quote_catalogue, quote_entry)
		);
		"""

		self.execute(sql)

		# if logging.root.level >= logging.DEBUG:
		# 	self.execute(f'PRAGMA table_info(quote);')
		# 	pprint(self.cursor.fetchall())

	def create_table_target(self):
		"""Create the target table"""
		logger.debug('HIPPO.Database.create_table_target()')
		sql = """CREATE TABLE target(
			target_id INTEGER PRIMARY KEY,
			target_name TEXT,
			target_metadata TEXT,
			CONSTRAINT UC_target UNIQUE (target_name)
		);
		"""

		self.execute(sql)

	def create_table_feature(self):
		"""Create the feature table"""
		logger.debug('HIPPO.Database.create_table_feature()')
		sql = """CREATE TABLE feature(
			feature_id INTEGER PRIMARY KEY,
			feature_family TEXT,
			feature_target INTEGER,
			feature_chain_name TEXT,
			feature_residue_name TEXT,
			feature_residue_number INTEGER,
			feature_atom_names TEXT,
			CONSTRAINT UC_feature UNIQUE (feature_family, feature_target, feature_chain_name, feature_residue_number, feature_residue_name, feature_atom_names)
		);
		"""

		self.execute(sql)

	def create_table_pattern_bfp(self):
		"""Create the pattern_bfp table"""
		logger.debug('HIPPO.Database.create_table_pattern_bfp()')

		sql = "CREATE VIRTUAL TABLE compound_pattern_bfp USING rdtree(compound_id, fp bits(2048))"
 
		self.execute(sql)

	def create_table_morgan_bfp(self):
		"""Create the morgan_bfp table"""
		logger.debug('HIPPO.Database.create_table_morgan_bfp()')
 
		sql = "CREATE VIRTUAL TABLE compound_morgan_bfp USING rdtree(compound_id, fp bits(2048))"
 
		self.execute(sql)

	### INSERTION

	def insert_compound(self, 
		*,
		smiles: str, 
		base: Compound | int | None = None, 
		alias: str | None = None,
		tags: None | list = None, 
		warn_duplicate=True,
		commit=True,
		metadata=None,
		inchikey: str = None, 
	) -> int:
		"""Insert a compound"""

		# process the base
		assert isinstance(base, int) or isinstance(base, Compound) or base is None, f'incompatible base={base}'

		if base and not isinstance(base,int):
			base = base.id

		# logger.debug(f'{base=}')

		# generate the inchikey name
		inchikey = inchikey or inchikey_from_smiles(smiles)

		# logger.debug(f'{smiles} --> {name}')

		sql = """
		INSERT INTO compound(compound_inchikey, compound_smiles, compound_base, compound_mol, compound_pattern_bfp, compound_morgan_bfp, compound_alias)
		VALUES(?1, ?2, ?3, mol_from_smiles(?2), mol_pattern_bfp(mol_from_smiles(?2), 2048), mol_morgan_bfp(mol_from_smiles(?2), 2, 2048), ?4)
		"""

		try:
			self.execute(sql, (inchikey, smiles, base, alias))

		except sqlite3.IntegrityError as e:
			if 'UNIQUE constraint failed: compound.compound_inchikey' in str(e):
				if warn_duplicate:
					logger.warning(f"Skipping compound with existing inchikey \"{inchikey}\"")
			elif 'UNIQUE constraint failed: compound.compound_smiles' in str(e):
				if warn_duplicate:
					logger.warning(f"Skipping compound with existing smiles \"{smiles}\"")
			elif 'UNIQUE constraint failed: compound.compound_pattern_bfp' in str(e):
				if warn_duplicate:
					logger.warning(f"Skipping compound with existing pattern binary fingerprint \"{smiles}\"")
			elif 'UNIQUE constraint failed: compound.compound_morgan_bfp' in str(e):
				if warn_duplicate:
					logger.warning(f"Skipping compound with existing morgan binary fingerprint \"{smiles}\"")
			else:
				logger.exception(e)

			# if tags:
				# logger.error('Unused tag data')
			return None

		except Exception as e:
			logger.exception(e)

		compound_id = self.cursor.lastrowid
		if commit:
			self.commit()

		### register the tags

		if tags:
			for tag in tags:
				self.insert_tag(name=tag, compound=compound_id, commit=commit)

		### register the binary fingerprints

		result = self.insert_compound_pattern_bfp(compound_id, commit=commit)

		if not result:
			logger.error('Could not insert compound pattern bfp')

		# ### register the morgan fingerprint

		# result = self.insert_compound_morgan_bfp(compound_id)

		# if not result:
		# 	logger.error('Could not insert compound morgan bfp')

		if metadata:
			self.insert_metadata(table='compound', id=compound_id, payload=metadata, commit=commit)

		return compound_id

	def insert_compound_pattern_bfp(self, compound_id, commit=True):
		"""Insert a compound_pattern_bfp"""

		sql = """
		INSERT INTO compound_pattern_bfp(compound_id, fp)
		VALUES(?1, ?2)
		"""
		
		bfp, = self.select_where('compound_pattern_bfp', 'compound', 'id', compound_id)

		try:
			self.execute(sql, (compound_id, bfp))

		except Exception as e:
			logger.exception(e)

		bfp_id = self.cursor.lastrowid
		if commit:
			self.commit()

		return bfp_id

	def insert_compound_morgan_bfp(self, compound_id):
		"""Insert a compound_morgan_bfp"""


		sql = """
		INSERT INTO compound_morgan_bfp(compound_id, fp)
		VALUES(?1, ?2)
		"""

		bfp, = self.select_where('compound_morgan_bfp', 'compound', 'id', compound_id)

		try:
			self.execute(sql, (compound_id, bfp))

		except Exception as e:
			logger.exception(e)

		mfp_id = self.cursor.lastrowid
		self.commit()

		return mfp_id

	def insert_pose(self,
		*,
		compound: Compound | int,
		target: int | str,
		path: str,
		inchikey: str | None = None,
		alias: str | None = None,
		reference: int | Pose | None = None,
		tags: None | list = None,
		energy_score: float | None = None,
		distance_score: float | None = None,
		metadata: None | dict = None,
		commit: bool = True,
		longname: str = None,
		warn_duplicate: bool = True,
		resolve_path: bool = True,
	):
		"""Insert a pose"""

		if isinstance(compound, Compound):
			compound = compound.id

		if isinstance(reference, Pose):
			reference = reference.id

		# assert isinstance(compound, Compound), f'incompatible {compound}'
		# assert isinstance(reference, Pose) or reference is None, f'incompatible reference={reference}'

		if isinstance(target, str):
			target = self.get_target_id(name=target)
			if not target:
				raise ValueError(f'No such {target=}')
		target_name = self.get_target_name(id=target)
		
		if resolve_path:
			try:
				path = Path(path)
				path = path.resolve(strict=True)
				path = str(path)

			except FileNotFoundError as e:
				logger.error(f'Path cannot be resolved: {mcol.file}{path}')
				return None

		sql = """
		INSERT INTO pose(pose_inchikey, pose_alias, pose_smiles, pose_compound, pose_target, pose_path, pose_reference, pose_energy_score, pose_distance_score)
		VALUES(?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9)
		"""

		try:
			self.execute(sql, (inchikey, alias, None, compound, target, path, reference, energy_score, distance_score))

		except sqlite3.IntegrityError as e:
			if 'UNIQUE constraint failed: pose.pose_path' in str(e):
				if warn_duplicate:
					logger.warning(f"Could not insert pose with duplicate path \"{path}\"")
			elif 'UNIQUE constraint failed: pose.pose_alias' in str(e):
				if warn_duplicate:
					logger.warning(f"Could not insert pose with duplicate alias \"{alias}\"")
			else:
				logger.exception(e)
			return None

		except Exception as e:
			logger.exception(e)
			raise
			
		pose_id = self.cursor.lastrowid
		if commit:
			self.commit()

		if tags:
			for tag in tags:
				self.insert_tag(name=tag, pose=pose_id, commit=commit)

		if metadata:
			self.insert_metadata(table='pose', id=pose_id, payload=metadata, commit=commit)

		return pose_id

	def insert_tag(self, 
		*,
		name, 
		compound: int = None, 
		pose: int = None,
		commit: bool = True,
	):
		"""Insert a tag"""

		assert compound or pose

		sql = """
		INSERT INTO tag(tag_name, tag_compound, tag_pose)
		VALUES(?1, ?2, ?3)
		"""

		try:
			self.execute(sql, (name, compound, pose))

		except sqlite3.IntegrityError as e:
			return None

		except Exception as e:
			logger.exception(e)		

		tag_id = self.cursor.lastrowid
		if commit:
			self.commit()
		return tag_id

	def insert_inspiration(self, 
		*,
		original: Pose | int, 
		derivative: Pose | int,
		warn_duplicate: bool = True,
		commit: bool = True,
	) -> int:
		"""Insert an inspiration"""

		if isinstance(original, Pose):
			original = original.id
		if isinstance(derivative, Pose):
			derivative = derivative.id

		assert isinstance(original, int), 'Must pass an integer ID or Pose object (original)'
		assert isinstance(derivative, int), 'Must pass an integer ID or Pose object (derivative)'

		sql = """
		INSERT INTO inspiration(inspiration_original, inspiration_derivative)
		VALUES(?1, ?2)
		"""

		try:
			self.execute(sql, (original, derivative))

		except sqlite3.IntegrityError as e:
			if warn_duplicate:
				logger.warning(f"Skipping existing inspiration: {original} {derivative}")
			return None

		except Exception as e:
			logger.exception(e)

		inspiration_id = self.cursor.lastrowid

		if commit:
			self.commit()
		return inspiration_id

	def insert_reaction(self,
		*,
		type: str,
		product: Compound | int,
		product_yield: float = 1.0,
		commit: bool = True,
	) -> int:
		"""Insert a reaction"""

		if isinstance(product, Compound):
			product = product.id

		# assert isinstance(product, Compound), f'incompatible {product=}'
		assert isinstance(type, str), f'incompatible {type=}'

		sql = """
		INSERT INTO reaction(reaction_type, reaction_product, reaction_product_yield)
		VALUES(?1, ?2, ?3)
		"""

		try:
			self.execute(sql, (type, product, product_yield))

		except Exception as e:
			logger.exception(e)
			
		reaction_id = self.cursor.lastrowid
		if commit:
			self.commit()
		return reaction_id

	def insert_reactant(self,
		*,
		compound: Compound,
		reaction: Reaction,
		amount: float = 1.0,
		commit: bool = True,
	) -> int:
		"""Insert a reactant"""

		if isinstance(reaction, int):
			reaction = self.get_reaction(id=reaction)

		if isinstance(compound, int):
			compound = self.get_compound(id=compound)

		assert isinstance(compound, Compound), f'incompatible {compound=}'
		assert isinstance(reaction, Reaction), f'incompatible {reaction=}'

		sql = """
		INSERT INTO reactant(reactant_amount, reactant_reaction, reactant_compound)
		VALUES(?1, ?2, ?3)
		"""

		try:
			self.execute(sql, (amount, reaction.id, compound.id))

		except sqlite3.IntegrityError as e:
			logger.warning(f"Skipping existing reactant: {reaction=} {compound=}")

		except Exception as e:
			logger.exception(e)
			
		reactant_id = self.cursor.lastrowid
		if commit:
			self.commit()
		return reactant_id

	def insert_quote(self,
		*,
		compound: Compound | int,
		supplier: str,
		catalogue: str | None,
		entry: str | None,
		amount: float,
		price: float,
		currency: str | None,
		purity: float | None,
		lead_time: int,
		smiles: str | None = None,
		commit: bool = True,
	) -> int | None:
		"""Insert a quote"""

		if not isinstance(compound, int):
			assert isinstance(compound, Compound), f'incompatible {compound=}'
			compound = compound.id
			
		assert currency in ['GBP', 'EUR', 'USD', None], f'incompatible {currency=}'
		
		assert supplier in ['MCule', 'Enamine', 'Stock'], f'incompatible {supplier=}'

		smiles = smiles or ""
		
		sql = """
		INSERT or REPLACE INTO quote(
			quote_smiles,
			quote_amount,
			quote_supplier,
			quote_catalogue,
			quote_entry,
			quote_lead_time,
			quote_price,
			quote_currency,
			quote_purity,
			quote_compound,
			quote_date
		)
		VALUES(?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, date());
		"""

		try:
			self.execute(sql, (smiles, amount, supplier, catalogue, entry, lead_time, price, currency, purity, compound))

		except sqlite3.InterfaceError as e:
			logger.error(e)
			logger.debug((smiles, amount, supplier, catalogue, entry, lead_time, price, currency, purity, compound))
			raise

		except Exception as e:
			logger.exception(e)
			return None
			
		quote_id = self.cursor.lastrowid
		if commit:
			self.commit()
		return quote_id

	def insert_target(self,
		*,
		name: str
	) -> int:
		"""Insert a target"""

		sql = """
		INSERT INTO target(target_name)
		VALUES(?1)
		"""

		try:
			self.execute(sql, (name, ))

		except sqlite3.IntegrityError as e:
			logger.warning(f"Skipping existing target with {name=}")
			return None

		except Exception as e:
			logger.exception(e)
			
		target_id = self.cursor.lastrowid
		self.commit()
		return target_id

	def insert_feature(self,
		*,
		family: str,
		target: int,
		chain_name: str,
		residue_name: str,
		residue_number: int,
		atom_names: list,
	) -> int:
		"""Insert a feature"""

		assert len(chain_name) == 1
		assert len(residue_name) <= 4
		for a in atom_names:
			assert len(a) <= 4

		if isinstance(target, str):
			target = self.get_target_id(name=target)
		assert isinstance(target, int)

		sql = """
		INSERT INTO feature(feature_family, feature_target, feature_chain_name, feature_residue_name, feature_residue_number, feature_atom_names)
		VALUES(?1, ?2, ?3, ?4, ?5, ?6)
		"""

		atom_names = ' '.join(atom_names)

		try:
			self.execute(sql, (family, target, chain_name, residue_name, residue_number, atom_names))

		except sqlite3.IntegrityError as e:
			return None

		except Exception as e:
			logger.exception(e)
			
		target_id = self.cursor.lastrowid
		self.commit()
		return target_id

	def insert_metadata(self,
		*,
		table: str,
		id: int,
		payload: dict,
		commit: bool = True,
	) -> None:
		"""Insert metadata"""

		payload = json.dumps(payload)

		self.update(table=table, id=id, key=f'{table}_metadata', value=payload, commit=commit)

	### SELECTION

	def select(self, query, table, multiple=False):
		"""Select entries"""

		sql = f'SELECT {query} FROM {table}'

		try:
			self.execute(sql)
		except sqlite3.OperationalError as e:
			logger.var('sql',sql)
			raise

		if multiple:
			result = self.cursor.fetchall()
		else:
			result = self.cursor.fetchone()

		return result

	def select_where(self, query, table, key, value=None, multiple=False, none='error', sort=None):
		"""Select entries where key==value"""

		if isinstance(value, str):
			value = f"'{value}'"

		if value is not None:
			where_str = f'{table}_{key}={value}'
		else:
			where_str = key

		if sort:
			sql = f'SELECT {query} FROM {table} WHERE {where_str} ORDER BY {sort}'
		else:
			sql = f'SELECT {query} FROM {table} WHERE {where_str}'

		try:
			self.execute(sql)
		except sqlite3.OperationalError as e:
			logger.var('sql',sql)
			raise

		if multiple:
			result = self.cursor.fetchall()
		else:
			result = self.cursor.fetchone()

		if not result and none == 'error':
			logger.error(f'No entry in {table} with {where_str}')
			return None

		return result

	def select_id_where(self, table, key, value, multiple=False, none='error'):
		"""Select ID's where key==value"""
		return self.select_where(query=f'{table}_id', table=table, key=key, value=value, multiple=multiple, none=none)

	def select_all_where(self, table, key, value, multiple=False, none='error'):
		"""Select * where key==value"""
		return self.select_where(query='*', table=table, key=key, value=value, multiple=multiple, none=none)

	### DELETION

	def delete_where(self, table, key, value=None):
		"""Delete where key==value"""

		if value is not None:

			if isinstance(value, str):
				value = f"'{value}'"

			sql = f'DELETE FROM {table} WHERE {table}_{key}={value}'

		else:
			
			sql = f'DELETE FROM {table} WHERE {key}'

		try:
			result = self.execute(sql)
			self.commit()

		except sqlite3.OperationalError as e:
			logger.var('sql',sql)
			raise

		return None

	def delete_tag(self, tag):
		self.delete_where(table='tag', key='name', value=tag)

	### UPDATE

	def update(self, *, 
		table: str, 
		id: int, 
		key: str, 
		value, 
		commit: bool = True,
	):
		"""Update a database entry where key==value"""

		# sql = f"""
		# UPDATE ?1
		# SET ?2 = ?3
		# WHERE ?4 = ?5;
		# """
		
		sql = f"""
		UPDATE {table}
		SET {key} = ?
		WHERE {table}_id = {id};
		"""
		
		try:
			# self.execute(sql, (table, key, value, f'{table}_id', id))
			self.execute(sql, (value, ))
		except sqlite3.OperationalError as e:
			logger.var('sql',sql)
			raise

		id = self.cursor.lastrowid
		if commit:
			self.commit()
		return id

	### GETTERS

	def get_compound(self,
		*,
		table: str = 'compound',
		id: int | None = None,
		inchikey: str | None = None,
		alias: str | None = None,
		smiles: str | None = None,
	) -> Compound:
		"""Get a compound"""
		
		if id is None:
			id = self.get_compound_id(inchikey=inchikey, smiles=smiles, alias=alias)

		if not id:
			logger.error(f'Invalid {id=}')
			return None

		query = 'compound_id, compound_inchikey, compound_alias, compound_smiles, compound_base'
		entry = self.select_where(query=query, table=table, key='id', value=id)
		compound = Compound(self._animal, self, *entry, metadata=None, mol=None)
		return compound

	def get_compound_id(self, 
		*,
		table: str = 'compound',
		inchikey: str | None = None, 
		alias: str | None = None, 
		smiles: str | None = None, 
	) -> int:
		"""Get a compound ID"""

		if inchikey:
			entry = self.select_id_where(table=table, key='inchikey', value=inchikey)

		elif alias:
			entry = self.select_id_where(table=table, key='alias', value=alias)

		elif smiles:
			entry = self.select_id_where(table=table, key='smiles', value=smiles)

		else:
			raise NotImplementedError

		if entry:
			return entry[0]

		return None

	def get_compound_computed_property(self, prop, compound_id):
		"""Use chemicalite to calculate a property from the stored binary molecule"""
		function = CHEMICALITE_COMPOUND_PROPERTY_MAP[prop]
		val, = self.select_where(query=f'{function}(compound_mol)', table='compound', key='id', value=compound_id, multiple=False)
		return val

	def get_pose(self,
		*,
		table: str = 'pose',
		id: int | None = None,
		inchikey: str = None,
		alias: str = None,
	) -> Pose:
		"""Get a pose"""
		
		if id is None:
			id = self.get_pose_id(inchikey=inchikey, alias=alias)

		if not id:
			logger.error(f'Invalid {id=}')
			return None

		query = 'pose_id, pose_inchikey, pose_alias, pose_smiles, pose_reference, pose_path, pose_compound, pose_target, pose_mol, pose_fingerprint, pose_energy_score, pose_distance_score'
		entry = self.select_where(query=query, table=table, key='id', value=id)
		pose = Pose(self, *entry)
		return pose

	def get_pose_id(self, 
		*,
		table: str = 'pose',
		inchikey: str | None = None, 
		alias: str | None = None, 
	) -> int:
		"""Get a pose ID"""

		if inchikey:
			# inchikey might not be unique
			entries = self.select_id_where(table=table, key='inchikey', value=inchikey, multiple=True)
			if len(entries) != 1:
				logger.warning(f'Multiple poses with {inchikey=}')
				return entries
			else:
				entry = entries[0]

		elif alias:
			entry = self.select_id_where(table=table, key='alias', value=alias)
		
		else:
			raise NotImplementedError
			
		if entry:
			return entry[0]

		return None

	def get_reaction(self,
		*,
		table: str = 'reaction',
		id: int | None = None,
		none: str | None = None,
	) -> Reaction:
		"""Get a reaction"""
		
		if not id:
			logger.error(f'Invalid {id=}')
			return None

		query = 'reaction_id, reaction_type, reaction_product, reaction_product_yield'
		entry = self.select_where(query=query, table=table, key='id', value=id, none=none)
		reaction = Reaction(self, *entry)
		return reaction

	def get_quote(self,
		*,
		table: str = 'quote',
		id: int | None = None,
		none: str | None = None,
	) -> list[dict]:
		"""Get a quote"""
		
		query = 'quote_compound, quote_supplier, quote_catalogue, quote_entry, quote_amount, quote_price, quote_currency, quote_lead_time, quote_purity, quote_date, quote_smiles, quote_id '
		entry = self.select_where(query=query, table=table, key='id', value=id, none=none)

		return Quote(
			db=self,
			id=entry[11],
			compound=entry[0],
			supplier=entry[1],
			catalogue=entry[2],
			entry=entry[3],
			amount=entry[4],
			price=entry[5],
			currency=entry[6],
			lead_time=entry[7],
			purity=entry[8],
			date=entry[9],
			smiles=entry[10],
		)

	def get_metadata(self, 
		*,
		table: str, 
		id: int
	) -> dict:
		"""Get metadata"""

		payload, = self.select_where(query=f'{table}_metadata', table=table, key=f'id', value=id)

		if payload:
			payload = json.loads(payload)
			
		else:
			payload = dict()

		metadata = MetaData(payload)

		metadata._db = self
		metadata._id = id
		metadata._table = table

		return metadata

	def get_target(self, 
		*,
		id=int
	) -> Target:
		"""Get target"""
		return Target(db=self,id=id,name=self.get_target_name(id=id))

	def get_target_name(self,
		*,
		id:int,
	) -> str:
		"""Get target name"""

		table = 'target'
		payload, = self.select_where(query=f'{table}_name', table=table, key='id', value=id)
		return payload

	def get_target_id(self, 
		*,
		name: str, 
	) -> int:
		"""Get target ID"""
			
		table = 'target'
		entry = self.select_id_where(table=table, key='name', value=name)

		if entry:
			return entry[0]

		return None

	def get_feature(self, 
		*,
		id=int
	) -> Feature:
		"""Get feature"""
		entry = self.select_all_where(table='feature', key='id', value=id)
		return Feature(*entry)

	def get_possible_reaction_ids(self, *, compound_ids):

		compound_ids_str = str(tuple(compound_ids)).replace(',)',')')

		result = self.execute(f'''
			WITH possible_reactants AS 
			(
				SELECT reactant_reaction, CASE WHEN reactant_compound IN {compound_ids_str} THEN reactant_compound END AS [possible_reactant] FROM reactant
			)

			, possible_reactions AS (
				SELECT reactant_reaction, COUNT(CASE WHEN possible_reactant IS NULL THEN 1 END) AS [count_null] FROM possible_reactants
				GROUP BY reactant_reaction
			)
			
			SELECT reactant_reaction FROM possible_reactions
			WHERE count_null = 0
		''').fetchall()

		return [q for q, in result]

	def get_unsolved_reaction_tree(self, *, product_ids, debug=False):

		from .cset import CompoundSet
		from .rset import ReactionSet

		all_reactants = set()
		all_reactions = set()

		print(product_ids)

		for product_id in product_ids:
			all_reactants.add(product_id)

		for i in range(300):

			if debug: 
				logger.var('recursive depth',i+1)
			
			if debug: 
				logger.var('#products', len(product_ids))

			product_ids_str = str(tuple(product_ids)).replace(',)',')')

			reaction_ids = self.select_where(table='reaction', query='DISTINCT reaction_id', key=f'reaction_product in {product_ids_str}', multiple=True, none='quiet')
			
			reaction_ids = [q for q, in reaction_ids]

			if not reaction_ids:
				break

			for reaction_id in reaction_ids:
				all_reactions.add(reaction_id)

			if debug: 
				logger.var('#reactions', len(reaction_ids))

			reaction_ids_str = str(tuple(reaction_ids)).replace(',)',')')

			reactant_ids = self.select_where(table='reactant', query='DISTINCT reactant_compound', key=f'reactant_reaction in {reaction_ids_str}', multiple=True)
			
			if debug: 
				logger.var('#reactants', len(reactant_ids))
			
			reactant_ids = [q for q, in reactant_ids]

			if not reactant_ids:
				break

			for reactant_id in reactant_ids:
				all_reactants.add(reactant_id)

			product_ids = reactant_ids

		# all intermediates
		ids = self.execute('SELECT DISTINCT reaction_product FROM reaction INNER JOIN reactant ON reaction.reaction_product = reactant.reactant_compound').fetchall()
		ids = [q for q, in ids]
		intermediates = CompoundSet(self, ids)

		# remove intermediates
		cset = CompoundSet(self, all_reactants)
		all_reactants = cset - intermediates

		# reactions
		all_reactions = ReactionSet(self, all_reactions)

		if debug:
			logger.var('#all_reactants', len(all_reactants))
			logger.var('#all_reactions', len(all_reactions))

		return all_reactants, all_reactions

	def get_reaction_price_estimate(self, *, reaction):

		# get reactants for a given reaction

		logger.warning('Price estimate does not account for branching!')
		reactants, _ = self.get_unsolved_reaction_tree(product_ids=reaction.reactant_ids)

		# how to make sure that there are no branching reactions?!

		# sum lowest unit price for each reactant

		price, = self.execute(f'''
		WITH unit_prices AS 
		(
		    SELECT quote_compound, MIN(quote_price/quote_amount) AS unit_price FROM quote 
		    WHERE quote_compound IN {reactants.str_ids}
		    GROUP BY quote_compound
		)
		SELECT SUM(unit_price) FROM unit_prices
		''').fetchone()

		return price

	def get_preferred_reaction_dict(self, *, compound_ids):

		print(compound_ids)

		raise NotImplementedError

	def get_possible_reaction_product_ids(self, *, reaction_ids):
		reaction_ids_str = str(tuple(reaction_ids)).replace(',)',')')
		return [q for q, in self.select_where(query='DISTINCT reaction_product', table='reaction', key=f'reaction_id IN {reaction_ids_str}', multiple=True)] 

	### COMPOUND QUERY

	def query_substructure(self, query, fast=True, none='error'):
		"""Search compounds by substructure"""

		# smiles
		if isinstance(query, str):

			if fast:
				sql = f"SELECT compound.compound_id FROM compound, compound_pattern_bfp AS bfp WHERE compound.compound_id = bfp.compound_id AND mol_is_substruct(compound.compound_mol, mol_from_smiles(?))"
			else:
				sql = f"SELECT compound_id FROM compound WHERE mol_is_substruct(compound_mol, mol_from_smiles(?))"

		else:

			raise NotImplementedError

		try:
			self.execute(sql, (query, ))
		except sqlite3.OperationalError as e:
			logger.var('sql',sql)
			raise

		result = self.cursor.fetchall()

		if not result and none == 'error':
			logger.error(f'No compounds with substructure {query}')
			return None

		compounds = []
		for id, in result:
			compounds.append(self.get_compound(id=id))

		return compounds

	def query_similarity(self, query, threshold, return_similarity=False, none='error'):
		"""Search compounds by similarity"""

		from .cset import CompoundSet

		# smiles
		if isinstance(query, str):

			if return_similarity:
				# sql = f"SELECT compound_id, bfp_tanimoto(mol_morgan_bfp(mol_from_smiles(?1), 2, 1024), mol_morgan_bfp(compound.compound_mol, 2, 1024)) as t FROM compound JOIN compound_morgan_bfp AS mfp USING(compound_id) WHERE mfp.compound_id match rdtree_tanimoto(mol_morgan_bfp(mol_from_smiles(?1), 2, 1024), ?2) ORDER BY t DESC "
				sql = f"SELECT compound_id, bfp_tanimoto(mol_pattern_bfp(mol_from_smiles(?1), 2048), mol_pattern_bfp(compound.compound_mol, 2048)) as t FROM compound JOIN compound_pattern_bfp AS mfp USING(compound_id) WHERE mfp.compound_id match rdtree_tanimoto(mol_pattern_bfp(mol_from_smiles(?1), 2048), ?2) ORDER BY t DESC "
			else:
				# sql = f"SELECT compound_id FROM compound_morgan_bfp AS mfp WHERE mfp.compound_id match rdtree_tanimoto(mol_morgan_bfp(mol_from_smiles(?1), 2, 1024), ?2) "
				sql = f"SELECT compound_id FROM compound_pattern_bfp AS mfp WHERE mfp.compound_id match rdtree_tanimoto(mol_pattern_bfp(mol_from_smiles(?1), 2048), ?2) "

		else:
			raise NotImplementedError

		try:
			self.execute(sql, (query, threshold))
		except sqlite3.OperationalError as e:
			logger.var('sql',sql)
			raise

		result = self.cursor.fetchall()

		if not result and none == 'error':
			logger.error(f'No compounds with similarity >= {threshold} to {query}')
			return None

		if return_similarity:
			ids, similarities = zip(*result)
			cset = CompoundSet(self, ids)
			return cset, similarities

		ids = [r for r, in result]
		cset = CompoundSet(self, ids)

		return cset

	def query_exact(self, query):
		"""Search for exact match compounds"""
		return self.query_similarity(query, 0.989, return_similarity=False)

	### COUNTING

	def count(self, table):
		"""Count all entries in a table"""

		sql = f"""SELECT COUNT(1) FROM {table}; """
		self.execute(sql)
		return self.cursor.fetchone()[0]

	def count_where(self, table, key, value=None):
		"""Count all entries in a table where key==value"""
		if value is not None:
			where_str = f'{table}_{key} is {value}'
		else:
			where_str = key

		sql = f"""SELECT COUNT(1) FROM {table} WHERE {where_str};"""
		self.execute(sql)
		return self.cursor.fetchone()[0]

	### ID SELECTION

	def min_id(self, table):
		"""Return the smallest ID in the given table"""
		id, = self.select(table=table, query=f'MIN({table}_id)')
		return id

	def max_id(self, table):
		"""Return the largest ID in the given table"""
		id, = self.select(table=table, query=f'MAX({table}_id)')
		return id

	def slice_ids(self, *, table, start, stop, step=1):
		"""Return a list of indices in the given slice"""

		min_id = self.min_id(table)
		max_id = self.max_id(table)

		start = start or min_id
		stop = stop or max_id + 1
		step = step or 1

		if not (start >= 0 and start <= max_id):
			raise IndexError(f'Slice {start=} outside of DB {table}_id range ({min_id}, {max_id})')

		if not (stop >= 0 and stop <= max_id + 1):
			raise IndexError(f'Slice {stop=} outside of DB {table}_id range ({min_id}, {max_id})')

		if step != 1:
			raise NotImplementedError(f'Slice {step=} not supported')

		ids = self.select_where(table=table, query=f'{table}_id', key=f'{table}_id >= {start} AND {table}_id < {stop}', multiple=True)
		ids = [q for q, in ids]

		return ids

	### PRUNING

	def prune_reactions(self, compound, reactions):
		"""Remove duplicate reactions for a given compound"""

		pruned = []
		del_list = []

		for i,reaction in enumerate(reactions):

			matches = [r for r in pruned if r == reaction]

			if not matches:
				pruned.append(reaction)

			else:
				del_list.append(reaction)

		for reaction in del_list:
			logger.warning(f'Deleted duplicate {reaction=}')
			self.delete_where('reaction', 'id', reaction.id)

		return pruned

	def remove_metadata_list_item(self, *, table, key, value, remove_empty=True):
		"""Remove a specific item from list-like values associated with a given key from all metadata entries in a given table"""

		# get id's with specific metadata key and value
		value_str = json.dumps(value)
		result = self.select_where(query=f'{table}_id, {table}_metadata', table=table, key=f'{table}_metadata LIKE \'%"export": [%{value_str}%]%\'', multiple=True, none='quiet')

		# loop over all matches
		for id, metadata_str in result:

			# read the metadata
			metadata = json.loads(metadata_str)
			
			# modify the metadata
			index = metadata[key].index(value)
			metadata[key].pop(index)

			if remove_empty and not metadata[key]:
				del metadata[key]
			
			# update the database
			metadata_str = json.dumps(metadata)
			self.update(table=table, id=id, key=f'{table}_metadata', value=metadata_str, commit=False)

		# only runs if non-zero matches
		else:
			# commit the changes
			self.commit()

	### PRINTING

	def list_tables(self):
		"""List all tables in the database"""
		self.execute("SELECT name FROM sqlite_master WHERE type='table';")
		pprint(self.cursor.fetchall())

	def print_table(self, name):
		"""Print a table's entries"""
		self.execute(f"SELECT * FROM {name}")
		pprint(self.cursor.fetchall())

	### DUNDERS

CHEMICALITE_COMPOUND_PROPERTY_MAP = {
	'num_heavy_atoms':'mol_num_hvyatms',
	'formula':'mol_formula',
	'num_rings':'mol_num_rings',
}




from .db import Database

import mcol

import os
from numpy import int64

import logging
logger = logging.getLogger('HIPPO')

from .reaction import Reaction

class ReactionTable:
	"""Object representing the 'reaction' table in the :class:`.Database`."""

	def __init__(self, 
		db: Database, 
		table: str = 'reaction',
	) -> None:
		
		self._db = db
		self._table = table

	### PROPERTIES

	@property
	def db(self):
		"""Returns the associated :class:`.Database`"""
		return self._db

	@property
	def table(self):
		return self._table

	@property
	def types(self):
		result = self.db.select(table=self.table, query='DISTINCT reaction_type', multiple=True)
		return [q for q, in result]

	@property
	def ids(self):
		"""Returns the IDs of child reactions"""
		result = self.db.select(table=self.table, query='reaction_id', multiple=True)
		return [q for q, in result]

	### METHODS

	def interactive(self):
		return self[self.ids].interactive()

	def get_by_type(self, reaction_type: str):
		result = self.db.select_where(table=self.table, query='reaction_id', key='type', value=reaction_type, multiple=True)
		return self[[q for q, in result]]

	def get_df(self, smiles=True, mols=True, **kwargs):
		"""Construct a pandas.DataFrame of all reactions in the database"""

		from rdkit.Chem import Mol
		from pandas import DataFrame

		### SQL QUERY

		data = {}
		
		if not smiles and not mols:
			
			sql = 'SELECT reaction_id, reaction_type, reaction_product, reactant_compound FROM reaction INNER JOIN reactant ON reaction.reaction_id = reactant.reactant_reaction'

			triples = self.db.execute(sql).fetchall()
			
			for reaction_id, product_id, reactant_id in triples:
				if reaction_id not in data:
					data[reaction_id] = dict(product_id=product_id, reactant_ids=[])
				else:
					assert data[reaction_id]['product_id'] == product_id

				data[reaction_id]['reactant_ids'].append(reactant_id)

		else:


			sql = '''
			SELECT {query}
			FROM reaction 

			INNER JOIN reactant 
			    ON reaction.reaction_id = reactant.reactant_reaction

			INNER JOIN compound c_r
			    ON c_r.compound_id = reactant.reactant_compound

			INNER JOIN compound c_p
			    ON c_p.compound_id = reaction.reaction_product
			'''

			if not mols:
				sql = sql.format(query='reaction_id, reaction_type, reaction_product, reactant_compound, c_p.compound_smiles, c_r.compound_smiles')

			else:
				sql = sql.format(query='reaction_id, reaction_type, reaction_product, reactant_compound, c_p.compound_smiles, c_r.compound_smiles, mol_to_binary_mol(c_p.compound_mol), mol_to_binary_mol(c_r.compound_mol)')

			results = self.db.execute(sql).fetchall()
			
			for result in results:

				reaction_id, reaction_type, product_id, reactant_id, product_smiles, reactant_smiles = result[:6]
				
				if mols:
					product_mol, reactant_mol = result[6:]

				if reaction_id not in data:
					data[reaction_id] = dict(reaction_id=reaction_id, reaction_type=reaction_type, product_id=product_id, reactant_ids=set(), product_smiles=product_smiles, reactant_smiles=set())
					if mols:
						data[reaction_id]['product_mol']=Mol(product_mol)
						data[reaction_id]['reactant_mols']=set()
				else:
					assert data[reaction_id]['product_id'] == product_id

				data[reaction_id]['reactant_ids'].add(reactant_id)
				data[reaction_id]['reactant_smiles'].add(reactant_smiles)
				if mols:
					data[reaction_id]['reactant_mols'].add(Mol(reactant_mol))

		data = data.values()
		return DataFrame(data)

	### DUNDERS

	def __getitem__(self, key) -> Reaction:
		
		match key:

			case int():

				if key == 0:
					return self.__getitem__(key=1)

				if key < 0:
					key = len(self) + 1 + key
					return self.__getitem__(key=key)

				else:
					return self.db.get_reaction(table=self.table, id=key)

			case key if isinstance(key, list) or isinstance(key, tuple) or isinstance(key, set):
				return ReactionSet(self.db, key)

			case slice():
				ids = self.db.slice_ids(table=self.table, start=key.start, stop=key.stop, step=key.step)
				return self[ids]

			case _:
				logger.error(f'Unsupported type for ReactionSet.__getitem__(): {key=} {type(key)}')

		return None

	def __repr__(self) -> str:
		return f'{mcol.bold}{mcol.underline}set(R x {len(self)}){mcol.unbold}{mcol.ununderline}'

	def __len__(self) -> int:
		return self.db.count(self.table)

	def __iter__(self):
		return iter(self[i+1] for i in range(len(self)))

class ReactionSet:
	"""Object representing a subset of the 'reaction' table in the :class:`.Database`."""
	
	def __init__(self,
		db: Database,
		indices: list = None,
		*,
		table: str = 'reaction',
	):
		
		self._db = db
		self._table = table
		indices = indices or []

		if not isinstance(indices, list):
			indices = list(indices)

		assert all(isinstance(i, int) or isinstance(i, int64) for i in indices)

		self._indices = sorted(list(set(indices)))

	### PROPERTIES

	@property
	def db(self):
		"""Returns the associated :class:`.Database`"""
		return self._db

	@property
	def table(self):
		return self._table

	@property
	def indices(self):
		return self._indices

	@property
	def ids(self):
		return self._indices

	@property
	def str_ids(self):
		return str(tuple(self.ids)).replace(',)',')')

	@property
	def products(self):
		"""Get all product compounds that can be synthesised with these reactions (no intermediates)"""
		from .cset import CompoundSet
		intermediates = self.intermediates
		product_ids = self.db.execute(f"""
			SELECT compound_id FROM compound
			INNER JOIN reaction ON compound_id = reaction_product
			WHERE reaction_id IN {self.str_ids}
			AND compound_id NOT IN {intermediates.str_ids}
		""").fetchall()
		return CompoundSet(self.db, [i for i, in product_ids])

	@property
	def intermediates(self):
		"""Get all intermediate compounds that can be synthesised with these reactions"""
		from .cset import CompoundSet
		sql = f"""
			SELECT DISTINCT compound_id FROM compound
			INNER JOIN reaction ON compound_id = reaction_product
			INNER JOIN reactant ON compound_id = reactant_compound
			WHERE reactant_reaction IN {self.str_ids}
		"""
		# print(sql)
		intermediate_ids = self.db.execute(sql).fetchall()
		return CompoundSet(self.db, [i for i, in intermediate_ids])

	### METHODS

	def add(self, r):
		assert r._table == 'reaction'
		if (id := r.id) not in self._indices:
			self._indices.append(id)

	def interactive(self):
		"""Creates a ipywidget to interactively navigate this PoseSet."""

		from ipywidgets import interactive, BoundedIntText, Checkbox, interactive_output, HBox, GridBox, Layout, VBox
		from IPython.display import display
		from pprint import pprint

		a = BoundedIntText(
				value=0,
				min=0,
				max=len(self)-1,
				step=1,
				description=f'Rs (/{len(self)}):',
				disabled=False,
			)

		b = Checkbox(description='Name', value=True)
		c = Checkbox(description='Summary', value=False)
		d = Checkbox(description='Draw', value=True)
		e = Checkbox(description='Check chemistry', value=False)
		f = Checkbox(description='Reactant Quotes', value=False)

		ui1 = GridBox([b, c, d], layout=Layout(grid_template_columns="repeat(5, 100px)"))
		ui2 = GridBox([e, f], layout=Layout(grid_template_columns="repeat(2, 150px)"))
		ui = VBox([a, ui1, ui2])
		
		def widget(i, name=True, summary=True, draw=True, check_chemistry=True, reactants=False):
			reaction = self[i]
			if name: print(repr(reaction))
			if summary: reaction.summary(draw=False)
			if draw: 
				reaction.draw()
			if check_chemistry: reaction.check_chemistry(debug=True)
			if reactants:
				for comp in reaction.reactants:
					# if summary:
						# comp.summary(draw=False)
					# elif name:
					print(repr(comp))

					quotes = comp.get_quotes(df=True)
					display(quotes)

					# break

					# if draw:
					# 	comp.draw()

		out = interactive_output(widget, {'i': a, 'name': b, 'summary': c, 'draw':d, 'check_chemistry':e, 'reactants':f })

		display(ui, out)
	
	def get_df(self, smiles=True, mols=True, **kwargs):
		"""Construct a pandas.DataFrame of this ReactionSet"""

		from pandas import DataFrame
		from tqdm import tqdm
		from rdkit.Chem import Mol

		logger.warning('Using slower Reaction.dict rather than direct SQL query')

		data = []
		for r in tqdm(self):
			data.append(r.get_dict(smiles=smiles, mols=mols, **kwargs))

		return DataFrame(data)

	def copy(self):
		return ReactionSet(self.db, self.ids)

	def get_recipes(self, amounts=1):

		from .recipe import Recipe
		return Recipe.from_reactions(db=self.db, reactions=self, amounts=1)

	### DUNDERS

	def __repr__(self) -> str:
		return f'{mcol.bold}{mcol.underline}''{'f'R x {len(self)}''}'f'{mcol.unbold}{mcol.ununderline}'

	def __len__(self) -> int:
		return len(self.indices)

	def __iter__(self):
		return iter(self.db.get_reaction(table=self.table, id=i) for i in self.indices)

	def __getitem__(self, key) -> Reaction:
		try:
			index = self.indices[key]
		except IndexError:
			logger.exception(f'list index out of range: {key=} for {self}')
			raise
		return self.db.get_reaction(table=self.table, id=index)

	def __add__(self, other):
		for reaction in other:
			self.add(reaction)
		return self
		




import mcol

from dataclasses import dataclass

from .target import Target

@dataclass
class Feature:
	"""Pharmocophoric feature in a protein"""

	id: int
	family: Target
	target: int
	chain_name: str
	residue_name: str
	residue_number: int
	atom_names: str

	def __str__(self):
		return f'F{self.id}'

	def __repr__(self):
		return f'{mcol.bold}{mcol.underline}{self.family} {self.chain_name} {self.residue_name} {self.residue_number} [{self.atom_names}]{mcol.unbold}{mcol.ununderline}'

	@property
	def chain_res_name_number_str(self):
		"""Return a string representation of the feature"""
		return f'{self.chain_name} {self.residue_name} {self.residue_number}'

	@property
	def res_name_number_str(self):
		"""Return a string representation of the feature"""
		return f'{self.residue_name} {self.residue_number}'





import pandas as pd
import numpy as np

from .cset import CompoundTable, IngredientSet, CompoundSet
from .pset import PoseTable, PoseSet
from .tags import TagTable
from .rset import ReactionTable
from .compound import Compound
from .reaction import Reaction
from .target import Target
from .pose import Pose

from .db import Database
from pathlib import Path

from tqdm import tqdm

from .tools import inchikey_from_smiles, sanitise_smiles, SanitisationError

from mlog import setup_logger
logger = setup_logger('HIPPO')

from rdkit.Chem import Mol

class HIPPO:

	"""The HIPPO 'animal' class. Instantiating a HIPPO object will create or link a HIPPO database.

	::
		
		from hippo import HIPPO
		animal = HIPPO(project_name, db_path)

	* *project_name* give this hippo a name
	* *db_path* path where the database will be stored


	"""
		
	def __init__(self, 
		name: str,  
		db_path: str | Path,
	):

		logger.header('Creating HIPPO animal')

		self._name = name
		# self._target_name = target

		logger.var('name', name, dict(color='arg'))
		# logger.var('target', target, dict(color='arg'))

		if not isinstance(db_path, Path):
			db_path = Path(db_path)
		
		logger.var('db_path', db_path, dict(color='file'))

		self._db_path = db_path
		self._db = Database(self.db_path, animal=self)

		self._compounds = CompoundTable(self.db, 'compound')
		self._poses = PoseTable(self.db, 'pose')
		self._tags = TagTable(self.db, 'tag')
		self._reactions = ReactionTable(self.db, 'reaction')

		### in memory subsets
		self._reactants = None
		self._products = None
		self._intermediates = None
		self._bases = None
		self._elabs = None

		logger.success(f"Initialised animal {self}")
		
	### FACTORIES

	### PROPERTIES

	@property
	def name(self) -> str:
		"""Returns the project name"""
		return self._name

	@property
	def db_path(self) -> str:
		"""Returns the database path"""
		return self._db_path

	@property
	def db(self):
		"""Returns the Database object"""
		return self._db

	@property
	def compounds(self):
		"""Returns a :doc:`CompoundTable <compounds>` object, which interfaces to the compound table in the database"""
		return self._compounds

	@property
	def poses(self):
		"""Returns a :doc:`PoseTable <poses>` object, which interfaces to the pose table in the database"""
		return self._poses

	@property
	def reactions(self):
		"""Returns a :doc:`ReactionSet <reactions>` object, which interfaces to the reaction table in the database"""
		return self._reactions

	@property
	def tags(self):
		"""Returns a :doc:`TagTable <metadata>` object, which interfaces to the tag table in the database"""
		return self._tags
	
	@property
	def num_compounds(self):
		"""Returns the number of compounds"""
		return len(self.compounds)

	@property
	def num_poses(self):
		"""Returns the number of poses"""
		return len(self.poses)

	@property
	def num_reactions(self):
		"""Returns the number of reactions"""
		return len(self.reactions)

	@property
	def num_tags(self):
		"""Returns the number of tags"""
		return len(self.tags.unique)

	@property
	def targets(self):
		"""Returns the targets registered in the DB"""
		target_ids = self.db.select(table='target', query='target_id', multiple=True)
		return [self.db.get_target(id=q) for q, in target_ids]

	@property
	def reactants(self):
		"""Returns a CompoundSet of all compounds that are used as a reactants"""
		if self._reactants is None or self._reactants['total_changes'] != self.db.total_changes:
			self._reactants = dict(set=self.compounds.reactants, total_changes=self.db.total_changes)
		return self._reactants['set']

	@property
	def products(self):
		"""Returns a CompoundSet of all compounds that are a product of a reaction but not a reactant"""
		if self._products is None or self._products['total_changes'] != self.db.total_changes:
			self._products = dict(set=self.compounds.products, total_changes=self.db.total_changes)
		return self._products['set']

	@property
	def intermediates(self):
		"""Returns a CompoundSet of all compounds that are products and reactants"""
		if self._intermediates is None or self._intermediates['total_changes'] != self.db.total_changes:
			self._intermediates = dict(set=self.compounds.intermediates, total_changes=self.db.total_changes)
		return self._intermediates['set']

	@property
	def num_reactants(self):
		"""Returns the number of reactants (see HIPPO.reactants)"""
		return len(self.reactants)

	@property
	def num_intermediates(self):
		"""Returns the number of intermediates (see HIPPO.intermediates)"""
		return len(self.intermediates)
	
	@property
	def num_products(self):
		"""Returns the number of products (see HIPPO.products)"""
		return len(self.products)

	@property
	def elabs(self):
		"""Returns a CompoundSet of all compounds that are a an elaboration of an existing base"""
		if self._elabs is None or self._elabs['total_changes'] != self.db.total_changes:
			self._elabs = dict(set=self.compounds.elabs, total_changes=self.db.total_changes)
		return self._elabs['set']

	@property
	def bases(self):
		"""Returns a CompoundSet of all compounds that are the basis for a set of elaborations"""
		if self._bases is None or self._bases['total_changes'] != self.db.total_changes:
			self._bases = dict(set=self.compounds.bases, total_changes=self.db.total_changes)
		return self._bases['set']

	@property
	def num_elabs(self):
		"""Returns the number of compounds that are a an elaboration of an existing base"""
		return len(self.elabs)

	@property
	def num_bases(self):
		"""Returns the number of compounds that are the basis for a set of elaborations"""
		return len(self.bases)

	### BULK INSERTION

	def add_hits(self, 
		target_name: str, 
		metadata_csv: str | Path, 
		aligned_directory: str | Path, 
		skip: list | None = None, 
		debug: bool = False, 
	) -> pd.DataFrame:

		"""Load in crystallographic hits downloaded from Fragalysis.

		* *target_name* Name of this protein target
		* *metadata_csv* Path to the metadata.csv from the Fragalysis download
		* *aligned_directory* Path to the aligned_files directory from the Fragalysis download
		* *skip* optional list of observation names to skip

		Returns: a DataFrame of metadata

		"""

		import molparse as mp
		from rdkit.Chem import PandasTools
		from .tools import remove_other_ligands

		if not isinstance(aligned_directory, Path):
			aligned_directory = Path(aligned_directory)

		# create the target
		target = self.register_target(name=target_name)

		skip = skip or []

		logger.var('aligned_directory',aligned_directory)

		count_directories_tried = 0
		count_compound_registered = 0
		count_poses_registered = 0

		meta_df = pd.read_csv(metadata_csv)
		generated_tag_cols = ['ConformerSites', 'CanonSites', 'CrystalformSites', 'Quatassemblies', 'Crystalforms']
		curated_tag_cols = [c for c in meta_df.columns if c not in ['Code', 'Long code', 'Compound code', 'Smiles', 'Downloaded']+generated_tag_cols]

		logger.var('curated_tag_cols',curated_tag_cols)

		for path in tqdm(list(aligned_directory.iterdir())):

			if not path.is_dir():
				continue

			if path.name in skip:
				continue

			count_directories_tried += 1

			sdfs = list(path.glob('*x?????.sdf'))

			assert len(sdfs) == 1, (path, sdfs)
			
			pdbs = list(path.glob('*x?????.pdb'))

			if not pdbs:
				pdbs = list(path.glob('*.pdb'))
				pdbs = [p for p in pdbs if '_ligand' not in p.name and '_apo' not in p.name and '_hippo' not in p.name]
				# observation_longname = pdbs[0].name.removesuffix('.pdb')
				# logger.var('observation_longname',observation_longname)
			# else:
				# observation_longname = None

			assert len(pdbs) == 1, (path, pdbs)
			
			# load the SDF
			df = PandasTools.LoadSDF(str(sdfs[0]), molColName='ROMol', idName='ID', strictParsing=True)

			# extract fields
			observation_shortname = path.name.replace('.sdf','')
			observation_longname = df.ID[0]
			mol = df.ROMol[0]
			lig_res_number = int(observation_longname.split('-')[1].split('_')[2])
			lig_chain = observation_longname.split('-')[1].split('_')[1]

			# smiles			
			smiles = mp.rdkit.mol_to_smiles(mol)
			smiles = sanitise_smiles(smiles, verbosity=debug)

			# parse PDB
			sys = mp.parse(pdbs[0], verbosity=debug)

			if len(sys.residues['LIG']) > 1 or any(r.contains_alternative_sites for r in sys.residues['LIG']):
				# create the single ligand bound pdb
				sys = remove_other_ligands(sys, lig_res_number, lig_chain)
				sys.prune_alternative_sites('A', verbosity=0)
				pose_path = str(pdbs[0].resolve()).replace('.pdb', '_hippo.pdb')
				mp.write(pose_path, sys, shift_name=True, verbosity=debug)
			else:
				pose_path = str(pdbs[0].resolve())
			
			# create the molecule / pose
			compound_id = self.db.insert_compound(
				# name=crystal_name, 
				smiles=smiles, 
				tags=['hits'],
				warn_duplicate=debug,
			)

			if not compound_id:

				inchikey = inchikey_from_smiles(smiles)
				compound = self.compounds[inchikey]

				if not compound:
					logger.error('Compound exists in database but could not be found by inchikey')
					logger.var('smiles',smiles)
					logger.var('inchikey',inchikey)
					logger.var('observation_shortname',observation_shortname)
					raise Exception
			
			else:
				count_compound_registered += 1
				compound = self.compounds[compound_id]

			# metadata


			meta_row = meta_df[meta_df['Code']==observation_shortname]
			if not len(meta_row):
				assert observation_longname
				meta_row = meta_df[meta_df['Long code']==observation_longname]

			# logger.var('observation_longname',observation_longname)

			assert len(meta_row)

			tags = ['hits']
			for tag in curated_tag_cols:
				if meta_row[tag].values[0]:
					tags.append(tag)

			metadata = {'observation_longname':observation_longname}
			for tag in generated_tag_cols:
				metadata[tag] = meta_row[tag].values[0]

			pose_id = self.db.insert_pose(
				compound=compound,
				alias=observation_shortname,
				target=target.id,
				path=pose_path,
				tags=tags,
				metadata=metadata,
			)

			if pose_id:
				count_poses_registered += 1

		logger.var('#directories parsed', count_directories_tried)
		logger.var('#compounds registered', count_compound_registered)
		logger.var('#poses registered', count_poses_registered)

	def add_compounds(self, 
		target: str,
		sdf_path: str | Path, 
		*,
		reference: int | Pose | None = None,
		tags: None | list = None,
		output_directory: str | Path | None = None,
		mol_col='ROMol',
		name_col='ID',
		inspiration_col = 'ref_mols',
		inspiration_map: None | dict = None,
		skip_first=False,
		convert_floats=True,
		stop_after: int | None = None,
		skip_equal_dict: dict | None = None,
		skip_not_equal_dict: dict | None = None,
		check_pose_RMSD: bool = False,
		pose_RMSD_tolerance: float = 1.0,
	):

		"""Add virtual hits from an SDF into the database.

		All columns except those specified via arguments mol_col and name_col are added to the Pose metadata.
		"""

		if not isinstance(sdf_path, Path):
			sdf_path = Path(sdf_path)

		logger.debug(f'{sdf_path=}')

		tags = tags or []

		from rdkit.Chem import PandasTools, MolToMolFile, MolFromMolFile
		from molparse.rdkit import mol_to_smiles, mol_to_pdb_block
		from numpy import isnan
		from pandas import read_pickle

		if sdf_path.name.endswith('.sdf'):
			df = PandasTools.LoadSDF(sdf_path)
		else:
			df = read_pickle(sdf_path)

		df_columns = list(df.columns)

		target = self.register_target(target).name

		assert mol_col in df_columns, f'{mol_col=} not in {df_columns}'
		assert name_col in df_columns, f'{name_col=} not in {df_columns}'
		assert inspiration_col in df_columns, f'{inspiration_col=} not in {df_columns}'

		df_columns.pop(df_columns.index(mol_col))
		df_columns.pop(df_columns.index(name_col))
		df_columns.pop(df_columns.index(inspiration_col))

		if not output_directory:
			import os
			output_directory = str(sdf_path.name).removesuffix('.sdf')
			logger.writing(f'Creating output directory {output_directory}')
			os.system(f'mkdir -p {output_directory}')

		output_directory = Path(output_directory)

		for i,row in tqdm(df.iterrows()):

			name = row[name_col].strip()

			if name == 'ver_1.2':
				logger.warning('Skipping Fragalysis header molecule')
				continue

			if skip_equal_dict and any(row[k] == v for k,v in skip_equal_dict.items()):
				continue

			if skip_not_equal_dict and any(row[k] != v for k,v in skip_not_equal_dict.items()):
				continue
			
			mol = row[mol_col]

			if not name:
				name = f'pose_{i}'

			mol_path = output_directory / f'{name}.mol'

			MolToMolFile(mol, mol_path)

			smiles = mol_to_smiles(mol)

			comp = self.register_compound(smiles=smiles, tags=tags)

			if not comp:
				logger.error(f'Could not register compound {i=}')
				continue

			# inspirations

			inspirations = []

			insp_str = row[inspiration_col]

			if isinstance(insp_str, str):
				insp_str = insp_str.removeprefix('[')
				insp_str = insp_str.removesuffix(']')
				insp_str = insp_str.replace("'", "")
				generator = insp_str.split(',')

			else:
				generator = insp_str

			for insp in generator:
				insp = insp.strip()

				if inspiration_map:
					if isinstance(inspiration_map, dict) and insp in inspiration_map:
						pose = inspiration_map[insp]
						if pose:
							inspirations.append(pose)
					elif hasattr(inspiration_map, '__call__'):
						pose = inspiration_map(insp)
						if pose:
							inspirations.append(pose)
					else:
						pose = self.poses[insp]
						if pose:
							inspirations.append(pose)
						else:
							logger.error(f'Could not find inspiration pose {insp}')
							continue
						
				else:
					pose = self.poses[insp]
					if pose:
						inspirations.append(pose)
					else:
						logger.error(f'Could not find inspiration pose {insp}')
						continue

			# metadata
			metadata = {}

			for col in df_columns:
				value = row[col]

				if not isinstance(col, str):
					if i == 0:
						logger.warning(f'Skipping metadata from column={col}.')
					continue

				if isinstance(value, float) and isnan(value):
					continue

				if convert_floats:
					try:
						value = float(value)
					except TypeError:
						pass
					except ValueError:
						pass

				if not (isinstance(value, str) or isinstance(value, float)):
					if i == 0:
						logger.warning(f'Skipping metadata from column={col}.')
					continue

				metadata[col] = value

			# print(metadata)

			pose = self.register_pose(
				alias=name, 
				compound=comp, 
				target=target, 
				path=mol_path, 
				metadata=metadata, 
				inspirations=inspirations, 
				tags=tags, 
				reference=reference,
				check_RMSD=check_pose_RMSD,
				RMSD_tolerance=pose_RMSD_tolerance,
			)

			if stop_after and i+1 >= stop_after:
				break

		logger.success(f'Loaded compounds from {sdf_path}')

	def add_syndirella_bases(self,
		df_path: str | Path,
		*,
		tags: None | list[str] = None,
		reaction_yield_map: dict | None = None,
		check_chemistry=True,
		debug=False,
	):

		from .chem import check_reaction_types, InvalidChemistryError

		tags = tags or []

		if isinstance(df_path, str):
			df_path = Path(df_path)
		df = pd.read_pickle(df_path)

		reaction_cols = [c for c in df.columns if '_reaction' in c]
		# print(reaction_cols)

		product_ids = set()

		# loop over rows

		for i,row in tqdm(df.iterrows()):

			# determine number of steps


			reaction_types = [r for c in reaction_cols if isinstance((r:=row[c]), str)]
			check_reaction_types(reaction_types)
			
			n_steps = len(reaction_types)

			if debug:
				logger.header(f'row={i}, {n_steps=}')

			try:
				
				# loop over each reaction step
				for j in range(n_steps):
	
					j += 1
	
					reactants = []
	
					reactant_previous_product = row[f'{j}_r_previous_product']
	
					# reactant 1
					if reactant_previous_product == 1:
						reactant1_id = product_id
						reactants.append(reactant1_id)
					elif smiles := row[f'{j}_r1_smiles']:
						
						if not isinstance(smiles, str):
							raise InvalidRowError(f'non-string {j}_r1_smiles')
													
						reactant1_id, duplicate = self.register_compound(smiles=smiles, commit=False, return_compound=False, tags=tags, base=None, register_base_if_duplicate=False, return_duplicate=True)
						
						# logger.var(f'reactant {j} 1', reactant1_id)
						
						reactants.append(reactant1_id)
				
					# reactant 2
					if reactant_previous_product == 2:
						reactant2_id = product_id
						reactants.append(reactant2_id)
					elif smiles := row[f'{j}_r2_smiles']:
						
						if not isinstance(smiles, str):
							raise InvalidRowError(f'non-string {j}_r2_smiles')

						reactant2_id, duplicate = self.register_compound(smiles=smiles, commit=False, return_compound=False, tags=tags, base=None, register_base_if_duplicate=False, return_duplicate=True)
						
						# logger.var(f'reactant {j} 2', reactant2_id)
							
						reactants.append(reactant2_id)
				
					# product
					if smiles := row[f'{j}_product_smiles']:
						if not isinstance(smiles, str):
							raise InvalidRowError(f'non-string {j}_product_smiles')
	
						if j != n_steps:
							this_tags = tags #+ ['intermediate']
	
						else:
							this_tags = ['Syndirella base'] + tags #+ ['Syndirella elaboration']
						
						product_id, duplicate = self.register_compound(smiles=smiles, tags=this_tags, commit=False, return_compound=False, base=None, register_base_if_duplicate=False, return_duplicate=True)
	
						if j == n_steps:
							product_ids.add(product_id)

						# logger.var(f'product {j}', product_id)

					# register the reaction
					if reaction_yield_map:
						product_yield = reaction_yield_map[row[f'{j}_reaction']]
					else:
						product_yield = 1.0
						
					try:

						reaction = self.register_reaction(
							reactants=reactants, 
							product=product_id, 
							type=row[f'{j}_reaction'], 
							commit=False, 
							product_yield=product_yield,
							check_chemistry=check_chemistry,
						)

						if debug:
							logger.var(str(reaction), reaction.reaction_str)

					except InvalidChemistryError as e:
						skipped_reactions += 1
			
			except InvalidRowError as e:
				# logger.error(f'Skipping invalid row {i=}: {e}')
				skipped_invalid_smiles += 1
				continue

			# chemistries = set(sum([df[df[f'{j+1}_reaction'].notnull()][f'{j+1}_reaction'].tolist() for j in range(n_steps)], []))
			# logger.var('Present reactions', str(reaction_types))

			# break

			# break

		return self.compounds[product_ids]

	def add_syndirella_elabs(self,
		df_path: str | Path,
		*,
		inspiration_map: dict | None = None,
		base_only: bool = False,
		tags: None | list[str] = None,
		reaction_yield_map: dict | None = None,
		require_truthy_bases: None | list[str] = None,
		require_truthy_elabs: None | list[str] = None,
		require_nonzero_truthy_bases: None | list[str] = None,
		stop_after=None,
		check_chemistry=True,
	):

		from .chem import check_reaction_types, InvalidChemistryError

		tags = tags or []
		require_truthy_bases = require_truthy_bases or ['path_to_mol', 'intra_geometry_pass']
		require_truthy_elabs = require_truthy_elabs or ['path_to_mol', 'intra_geometry_pass']
		require_nonzero_truthy_bases = require_nonzero_truthy_bases or ['path_to_mol', 'intra_geometry_pass']
		assert all([k in require_truthy_bases for k in require_nonzero_truthy_bases])
		
		if isinstance(df_path, str):
			df_path = Path(df_path)
		df = pd.read_pickle(df_path)

		# work out number of reaction steps
		n_steps = max([int(s.split('_')[0]) for s in df.columns if '_product_smiles' in s])

		# check chemistries
		chemistries = set(sum([df[df[f'{j+1}_reaction'].notnull()][f'{j+1}_reaction'].tolist() for j in range(n_steps)], []))
		logger.var('Present reactions', str(chemistries))
		check_reaction_types(chemistries)

		if f'{n_steps}_num_atom_diff' not in df.columns:
			logger.error(df_path)
			logger.error(f'{n_steps}_num_atom_diff not in columns:')
			print(df.columns)
			raise NotImplementedError

		for key in require_nonzero_truthy_bases:
			if not any(df[df[f'{n_steps}_product_name'].str.contains("base")][key].values):
				logger.warning(f'No bases have {key}. Inserting them anyway')
				require_truthy_bases.pop(require_truthy_bases.index(key))

		base_id = None
		base_reactants = {}

		if base_only:
			generator = df.iterrows()
		else:
			generator = tqdm(df.iterrows())

		n_comps = len(self.compounds)
		n_poses = len(self.poses)
		n_reactions = len(self.reactions)

		skipped_smaller = 0
		skipped_reactions = 0
		skipped_invalid_smiles = 0

		try:

			for i,row in generator:

				path_to_mol = row.path_to_mol

				this_row_is_a_base = 'base' in row[f'{n_steps}_product_name']

				# skip entries that have non-truthy columns
				if this_row_is_a_base:

					if any(not row[key] for key in require_truthy_bases):
						continue

					# for key in warn_not_truthy_bases:
					# 	if not row[key]:
					# 		logger.warning(f'Base (row {i=}) has {key}={row[key]}')

				elif any(not row[key] for key in require_truthy_elabs):
					continue
				
				if row[f'{n_steps}_num_atom_diff'] <= 0 and not this_row_is_a_base:
					skipped_smaller += 1
					continue

				if base_only and not this_row_is_a_base:
					continue

				if base_only and base_id and not this_row_is_a_base:
					break

				if this_row_is_a_base and skipped_smaller and not base_id:
					logger.warning(f"Skipped {skipped_smaller} elaborations that are smaller than the base compound")

				elabs_registered = set()

				try:
					
					# loop over each reaction step
					for j in range(n_steps):
		
						j += 1
		
						reactants = []
		
						reactant_previous_product = row[f'{j}_r_previous_product']
		
						# reactant 1
						if reactant_previous_product == 1:
							reactant1_id = product_id
							reactants.append(reactant1_id)
						elif smiles := row[f'{j}_r1_smiles']:
							
							if not isinstance(smiles, str):
								raise InvalidRowError(f'non-string {j}_r1_smiles')
							
							try:
								base = base_reactants[j][1] if not this_row_is_a_base else None
							except KeyError:
								print(row)
								logger.error(f'Expected base_reactants to contain data when {this_row_is_a_base=}')
								raise
								
							reactant1_id, duplicate = self.register_compound(smiles=smiles, commit=False, return_compound=False, tags=tags, base=base, register_base_if_duplicate=False, return_duplicate=True)

							if duplicate and reactant1_id not in elabs_registered:
								if base:
									self.db.update(table='compound', id=reactant1_id, key='compound_base', value=base, commit=False)
								elabs_registered.add(reactant1_id)
							
							reactants.append(reactant1_id)
					
						# reactant 2
						if reactant_previous_product == 2:
							reactant2_id = product_id
							reactants.append(reactant2_id)
						elif smiles := row[f'{j}_r2_smiles']:
							
							if not isinstance(smiles, str):
								raise InvalidRowError(f'non-string {j}_r2_smiles')

							# base = base_reactants[j][1] if not this_row_is_a_base else None
							# reactant2_id = self.register_compound(smiles=smiles, commit=False, return_compound=False, tags=tags)
							base = base_reactants[j][2] if not this_row_is_a_base else None

							reactant2_id, duplicate = self.register_compound(smiles=smiles, commit=False, return_compound=False, tags=tags, base=base, register_base_if_duplicate=False, return_duplicate=True)

							if duplicate and reactant2_id not in elabs_registered:
								if base:
									self.db.update(table='compound', id=reactant2_id, key='compound_base', value=base, commit=False)
								elabs_registered.add(reactant2_id)
								
							reactants.append(reactant2_id)
					
						# product
						if smiles := row[f'{j}_product_smiles']:
							if not isinstance(smiles, str):
								raise InvalidRowError(f'non-string {j}_product_smiles')
		
							if j != n_steps:
								this_tags = tags #+ ['intermediate']
								# base = base_reactants[j]['product'] if not this_row_is_a_base else None
								base = None
		
							elif this_row_is_a_base:
								this_tags = ['Syndirella base'] + tags
								base = None
		
							else:
								this_tags = tags #+ ['Syndirella elaboration']
								base = base_id

							product_id, duplicate = self.register_compound(smiles=smiles, tags=this_tags, commit=False, return_compound=False, base=base, register_base_if_duplicate=False, return_duplicate=True)

							if duplicate and product_id not in elabs_registered:
								if base:
									self.db.update(table='compound', id=product_id, key='compound_base', value=base, commit=False)
								elabs_registered.add(product_id)

							if not base_id:
								base_reactants[j] = {1:reactant1_id, 2:reactant2_id, 'product':product_id }
							
							if not base_id and j == n_steps and this_row_is_a_base:
								base_id = product_id
		
						# register the reaction
						if reaction_yield_map:
							product_yield = reaction_yield_map[row[f'{j}_reaction']]
						else:
							product_yield = 1.0
							
						try:
							self.register_reaction(
								reactants=reactants, 
								product=product_id, 
								type=row[f'{j}_reaction'], 
								commit=False, 
								product_yield=product_yield,
								check_chemistry=check_chemistry,
							)
						except InvalidChemistryError as e:
							skipped_reactions += 1
				
				except InvalidRowError as e:
					# logger.error(f'Skipping invalid row {i=}: {e}')
					skipped_invalid_smiles += 1
					continue

				# pose metadata
				metadata = {}

				# inspirations
				inspirations = []
				for inspiration in row.regarded:
					if inspiration_map:
						inspiration = inspiration_map[inspiration]
					else:
						# this is really expensive
						inspiration = self.poses[inspiration]
					if inspiration:
						inspirations.append(inspiration)
				
				if path_to_mol:
					
					path_to_mol = path_to_mol.replace('//','/')

					# register the pose
					self.register_pose(
						compound=product_id, 
						target='A71EV2A', 
						path=path_to_mol, 
						inspirations=inspirations, 
						metadata=metadata, 
						tags=this_tags, 
						commit=False, 
						return_pose=False,
						overwrite_metadata=True,
						warn_duplicate=False,
						energy_score=row['∆∆G'],
						distance_score=row['comRMSD'],
					)

				if stop_after and i > stop_after:
					break

		except KeyboardInterrupt:
			logger.error('KeyboardInterrupt')

		self.db.commit()

		n_comps = len(self.compounds) - n_comps
		n_poses = len(self.poses) - n_poses
		n_reactions = len(self.reactions) - n_reactions

		if skipped_reactions:
			logger.warning(f'Skipped {skipped_reactions} invalid reactions')

		if skipped_invalid_smiles:
			logger.warning(f'Skipped {skipped_invalid_smiles} rows with NaN smiles')

		if n_comps:
			if not base_only: 
				logger.success(f'Loaded {n_comps} new compounds from {df_path.name}')
		else:
			logger.warning(f'Loaded {n_comps} new compounds from {df_path.name}')
		
		if n_poses:
			if not base_only: 
				logger.success(f'Loaded {n_poses} new poses from {df_path.name}')
		else:
			logger.warning(f'Loaded {n_poses} new poses from {df_path.name}')

		if n_reactions:
			if not base_only: 
				logger.success(f'Loaded {n_reactions} new reactions from {df_path.name}')
		else:
			logger.warning(f'Loaded {n_reactions} new reactions from {df_path.name}')

		# return base_id

	def add_enamine_quote(
		self, 
		path: str | Path, 
		*, 
		orig_name_col: str = 'Diamond ID (Molecule Name)',
		entry_col = 'Catalog ID',
	):

		"""Load an Enamine quote provided as an excel file"""

		df = pd.read_excel(path)

		smiles_col = 'SMILES'
		purity_col = 'Purity, %'
		amount_col = 'Amount, mg'
		catalogue_col = 'Collection'
		lead_time_col = 'Lead time'

		assert smiles_col in df.columns, 'Unexpected Excel format (smiles_col)'
		assert orig_name_col in df.columns, 'Unexpected Excel format (orig_name_col)'
		assert entry_col in df.columns, 'Unexpected Excel format (entry_col)'
		assert purity_col in df.columns, 'Unexpected Excel format (purity_col)'
		assert amount_col in df.columns, 'Unexpected Excel format (amount_col)'
		assert catalogue_col in df.columns, 'Unexpected Excel format (catalogue_col)'
		assert 'Price, EUR' in df.columns or 'Price, USD' in df.columns, 'Unexpected Excel format (price_col)'
		assert lead_time_col in df.columns, 'Unexpected Excel format (lead_time_col)'

		price_cols = [c for c in df.columns if c.startswith('Price')]
		assert len(price_cols) == 1
		price_col = price_cols[0]
		currency = price_col.split(', ')[-1]

		ingredients = IngredientSet(self.db)
	
		for i,row in df.iterrows():
			smiles = row[smiles_col]

			if not isinstance(smiles, str):
				break

			compound = self.register_compound(smiles=smiles)


			if (catalogue := row[catalogue_col]) == 'No starting material':
				continue

			if (price := row[price_col]) == 0.0:
				continue

			if not isinstance(row[lead_time_col], str):
				continue

			if 'week' in row[lead_time_col]:
				lead_time = int(row[lead_time_col].split()[0].split('-')[-1])*5
			else:
				raise NotImplementedError

			quote_data = dict(
				compound=compound,
				supplier='Enamine',
				catalogue=catalogue,
				entry=row[entry_col],
				amount=row[amount_col],
				purity=row[purity_col]/100,
				lead_time=lead_time,
				price=price,
				currency=currency,
				smiles=smiles,
			)

			q_id = self.db.insert_quote(**quote_data)

			ingredients.add(
				compound_id=compound.id,
				amount=row[amount_col],
				quote_id=q_id,
				supplier='Enamine',
				max_lead_time=None,
			)

		return ingredients

	def add_mcule_quote(
		self,
		path: str | Path,
	):

		### get lead time from suppliers sheet

		sheet_name: str = 'List of suppliers'
		df = pd.read_excel(path, sheet_name=sheet_name)

		supplier_col = 'Supplier'
		lead_time_col = 'Delivery time (working days)'

		assert supplier_col in df.columns, 'Unexpected Excel format (supplier_col)'
		assert lead_time_col in df.columns, 'Unexpected Excel format (lead_time_col)'

		lead_time_lookup = { row[supplier_col]:row[lead_time_col] for i,row in df.iterrows() }

		### parse individual compound quotes

		sheet_name: str = 'List of products'
		df = pd.read_excel(path, sheet_name=sheet_name)

		# return df

		smiles_col = 'Quoted product SMILES'
		entry_col = 'Query Mcule ID'
		purity_col = 'Guaranteed purity (%)'
		amount_col = 'Quoted Amount (mg)'
		catalogue_col = 'Supplier'
		lead_time_col = 'Lead time'
		price_col = 'Product price (USD)'
		currency = 'USD'

		assert smiles_col in df.columns, 'Unexpected Excel format (smiles_col)'
		assert entry_col in df.columns, 'Unexpected Excel format (entry_col)'
		assert purity_col in df.columns, 'Unexpected Excel format (purity_col)'
		assert amount_col in df.columns, 'Unexpected Excel format (amount_col)'
		assert catalogue_col in df.columns, 'Unexpected Excel format (catalogue_col)'
		assert price_col in df.columns, 'Unexpected Excel format (price_col)'

		ingredients = IngredientSet(self.db)
	
		for i,row in tqdm(df.iterrows()):
			smiles = row[smiles_col]

			if not isinstance(smiles, str):
				break

			compound = self.register_compound(smiles=smiles)

			# if (catalogue := row[catalogue_col]) == 'No starting material':
			# 	continue

			catalogue = row[catalogue_col]
			lead_time = lead_time_lookup[catalogue]

			# if (price := row[price_col]) == 0.0:
			# 	continue

			# if not isinstance(row[lead_time_col], str):
				# continue

			# if 'week' in row[lead_time_col]:
			# 	lead_time = int(row[lead_time_col].split()[0].split('-')[-1])*5
			# else:
			# 	raise NotImplementedError

			quote_data = dict(
				compound=compound,
				supplier='MCule',
				catalogue=catalogue,
				entry=row[entry_col],
				amount=row[amount_col],
				purity=row[purity_col]/100,
				lead_time=lead_time,
				price=row[price_col],
				currency=currency,
				smiles=smiles,
			)

			q_id = self.db.insert_quote(**quote_data, commit=False)

			ingredients.add(
				compound_id=compound.id,
				amount=row[amount_col],
				quote_id=q_id,
				supplier='MCule',
				max_lead_time=None,
			)

		self.db.commit()

		return ingredients

	### SINGLE INSERTION

	def register_compound(self, 
		*,
		smiles: str, 
		base: Compound | int | None = None, 
		tags: None | list = None, 
		metadata: None | dict = None,
		return_compound: bool = True,
		commit: bool = True,
		alias: str | None = None,
		return_duplicate: bool = False,
		register_base_if_duplicate: bool = True,
		radical: str = 'warning',
	) -> Compound:

		"""Use a smiles string to add a compound to the database. If it already exists return the compound"""

		assert smiles
		assert isinstance(smiles, str), f'Non-string {smiles=}'
		
		try:
			smiles = sanitise_smiles(smiles, sanitisation_failed='error', radical=radical)
		except SanitisationError as e:
			logger.error(f'Could not sanitise {smiles=}')
			logger.error(str(e))
			return None
		except AssertionError:
			logger.error(f'Could not sanitise {smiles=}')
			return None

		if base and isinstance(base, Compound):
			base = base.id

		inchikey = inchikey_from_smiles(smiles)

		compound_id = self.db.insert_compound(
			smiles=smiles, base=base, inchikey=inchikey, tags=tags, 
			metadata=metadata, warn_duplicate=False, commit=commit, alias=alias)

		duplicate = not bool(compound_id)

		def _return(compound, duplicate, return_compound, return_duplicate):
			if not return_compound and not isinstance(compound, int):
				compound = compound.id
			if return_duplicate:
				return compound, duplicate
			else:
				return compound

		if return_compound or metadata or alias or tags:
			if not compound_id:
				compound = self.compounds[inchikey]
			else:
				compound = self.compounds[compound_id]

			if metadata:
				compound.metadata.update(metadata)

			if alias:
				compound.alias = alias

			if base and not (register_base_if_duplicate and duplicate):
				compound.base = base

			if tags:
				for tag in tags:
					compound.tags.add(tag)
			
			return _return(compound, duplicate, return_compound, return_duplicate)
			
			if return_compound:
				return compound
			else:
				return compound.id

		else:
			if not compound_id:
				compound_id = self.db.get_compound_id(inchikey=inchikey)

				if base and not (register_base_if_duplicate and duplicate):
					self.db.update(table='compound', id=compound_id, key='compound_base', value=base, commit=commit)
			
			return _return(compound_id, duplicate, return_compound, return_duplicate)

	def register_reaction(self,
		*,
		type: str,
		product: Compound | int,
		reactants: list[Compound | int],
		commit: bool = True,
		product_yield: float = 1.0,
		check_chemistry: bool = False,
	) -> Reaction:

		### CHECK REACTION VALIDITY

		if check_chemistry:
			from .chem import check_chemistry, InvalidChemistryError

			if not isinstance(product, Compound):
				product = self.db.get_compound(id=product)

			if not isinstance(reactants, CompoundSet):
				reactants = CompoundSet(self.db, reactants)

			valid = check_chemistry(type, reactants, product)

			if not valid:
				raise InvalidChemistryError(f'{type=}, {reactants.ids=}, {product.id=}')

		### CHECK FOR DUPLICATES
		
		if isinstance(product, Compound):
			product = product.id

		reactant_ids = set(v.id if isinstance(v, Compound) else v for v in reactants)

		pairs = self.db.execute(f'SELECT reactant_reaction, reactant_compound FROM reactant INNER JOIN reaction ON reactant.reactant_reaction = reaction.reaction_id WHERE reaction_type="{type}" AND reaction_product = {product}').fetchall()

		if pairs:

			reax_dict = {}
			for reaction_id, reactant_id in pairs:
				if reaction_id not in reax_dict:
					reax_dict[reaction_id] = set()
				reax_dict[reaction_id].add(reactant_id)

			for reaction_id, reactants in reax_dict.items():
				if reactants == reactant_ids:
					return self.reactions[reaction_id]
				
		### INSERT A NEW REACTION

		reaction_id = self.db.insert_reaction(type=type, product=product, commit=commit, product_yield=product_yield)
		
		### INSERT REACTANTS

		for reactant in reactant_ids:
			self.db.insert_reactant(compound=reactant, reaction=reaction_id, commit=commit)

		return self.reactions[reaction_id]

	def register_target(self,
		name: str,
	) -> Target:
		
		target_id = self.db.insert_target(name=name)

		if not target_id:
			target_id = self.db.get_target_id(name=name)

		return self.db.get_target(id=target_id)

	def register_pose(self,
		*,
		compound: Compound | int,
		target: str,
		path: str,
		inchikey: str | None = None,
		alias: str | None = None,
		reference: int | None = None,
		tags: None | list = None,
		metadata: None | dict = None,
		inspirations: None | list = None,
		return_pose: bool = True,
		energy_score: float | None = None,
		distance_score: float | None = None,
		commit: bool = True,
		overwrite_metadata: bool = True,
		warn_duplicate: bool = True,
		check_RMSD: bool = False,
		RMSD_tolerance: float = 1.0,
		split_PDB: bool = False,
		template_mol: str | Mol | None = None,
		duplicate_alias: str = 'modify',
	) -> Pose:

		assert duplicate_alias in ['error', 'modify']

		from molparse import parse

		if split_PDB:

			sys = parse(path, verbosity=False, alternative_site_warnings=False)

			lig_residues = []

			for res in sys.ligand_residues:
				lig_residues += res.split_by_site()
			
			if len(lig_residues) > 1:

				assert not energy_score
				assert not distance_score

				logger.warning(f'Splitting ligands in PDB: {path}')

				results = []
				for i,res in enumerate(lig_residues):
					file = str(path).replace('.pdb',f'_hippo_{i}.pdb')

					split_sys = sys.protein_system

					for atom in res.atoms:
						split_sys.add_atom(atom)

					logger.writing(file)
					split_sys.write(file, verbosity=False)

					result = self.register_pose(
						compound = compound,
						target = target,
						path = file,
						inchikey = inchikey,
						alias = alias,
						reference = reference,
						tags = tags,
						metadata = metadata,
						inspirations = inspirations,
						return_pose = return_pose,
						commit = commit,
						overwrite_metadata = overwrite_metadata,
						warn_duplicate = warn_duplicate,
						check_RMSD = check_RMSD,
						RMSD_tolerance = RMSD_tolerance,
						split_PDB = False,
						template_mol = template_mol,
					)

					results.append(result)

				return results

		if template_mol:
			raise NotImplementedError

		if isinstance(compound, int):
			compound_id = compound
		else:
			compound_id = compound.id

		if check_RMSD:
			
			# check if the compound has existing poses
			other_pose_ids = self.db.select_id_where(table='pose', key='compound', value=compound_id, none='quiet', multiple=True)
			
			if other_pose_ids:
				other_poses = PoseSet(self.db, [i for i, in other_pose_ids])

				from molparse.rdkit import draw_mols, draw_flat
				from rdkit.Chem import MolFromMolFile
				from numpy.linalg import norm
				from numpy import array
				mol = MolFromMolFile(str(path.resolve()))
				
				c1 = mol.GetConformer()
				atoms1 = [a for a in mol.GetAtoms()]
				symbols1 = [a.GetSymbol() for a in atoms1]
				positions1 = [c1.GetAtomPosition(i) for i,_ in enumerate(atoms1)]

				for pose in other_poses:
					c2 = pose.mol.GetConformer()
					atoms2 = [a for a in pose.mol.GetAtoms()]
					symbols2 = [a.GetSymbol() for a in atoms2]
					positions2 = [c2.GetAtomPosition(i) for i,_ in enumerate(atoms2)]
					
					for s1,p1 in zip(symbols1, positions1):
						for s2,p2 in zip(symbols2, positions2):
							if s2 != s1:
								continue
							if norm(array(p2-p1)) <= RMSD_tolerance:
								# this atom (1) is within tolerance
								break
						else:
							# this atom (1) is outside of tolerance
							break
					else:
						# all atoms within tolerance --> too similar
						logger.warning(f'Found similar {pose=}')
						if return_pose:
							return pose
						else:
							return pose.id

		pose_data = dict(
			compound=compound, 
			inchikey=inchikey, 
			alias=alias, 
			target=target, 
			path=path, 
			tags=tags, 
			metadata=metadata, 
			reference=reference, 
			warn_duplicate=warn_duplicate, 
			commit=commit, 
			energy_score=energy_score, 
			distance_score=distance_score,
		)
					
		pose_id = self.db.insert_pose(**pose_data)
		
		if not pose_id:
			
			# constraint failed
			if isinstance(path, Path):
				path = path.resolve()

			# try getting by path
			result = self.db.select_where(table='pose', query='pose_id', key='path', value=str(path), none='quiet')

			# try getting by alias
			if not result:
				result = self.db.select_where(table='pose', query='pose_id', key='alias', value=alias)

				if result and duplicate_alias == 'error':
					raise Exception('could not register pose with existing alias')

				elif result and duplicate_alias == 'modify':

					new_alias = alias + '_copy'

					logger.warning(f'Modifying alias={alias} --> {new_alias}')

					pose_data['alias'] = new_alias
					pose_id = self.db.insert_pose(**pose_data)

				else:
					pose_id, = result
			else:
				pose_id, = result

			assert pose_id

		if not pose_id:
			logger.var('compound', compound)
			logger.var('inchikey', inchikey)
			logger.var('alias', alias)
			logger.var('target', target)
			logger.var('path', path)
			logger.var('reference', reference)
			logger.var('tags', tags)
			logger.debug(f'{metadata=}')
			logger.debug(f'{inspirations=}')

			raise Exception

		if return_pose or (metadata and not overwrite_metadata):
			pose = self.poses[pose_id]

			if metadata:
				pose.metadata.update(metadata)
		
		else:
			pose = pose_id

		if overwrite_metadata:
			self.db.insert_metadata(table='pose', id=compound_id, payload=metadata, commit=commit)

		inspirations = inspirations or []
		for inspiration in inspirations:
			self.db.insert_inspiration(original=inspiration, derivative=pose, warn_duplicate=False, commit=commit)

		return pose

	### QUOTING

	def quote_compounds(self, quoter, compounds):
		"""Get batch quotes using the hippo.Quoter object supplied and add the quotes to the database"""
		logger.header(f'Getting {quoter.supplier} quotes for {len(compounds)} compounds')
		batch_size = quoter.batch_size
		for i in range(0, len(compounds), batch_size):
			logger.debug(f'batch {i%batch_size}')
			batch = compounds[i:i+batch_size]
			quoter.get_batch_quote(batch)

	def quote_reactants(self, quoter, *, unquoted_only=False):
		"""Get batch quotes for all reactants in the database"""

		if unquoted_only:
			compounds = self.reactants.get_unquoted(supplier=quoter.supplier)
		else:
			compounds = self.reactants

		self.quote_compounds(quoter=quoter, compounds=compounds)
	
	def quote_intermediates(self, quoter):
		"""Get batch quotes for all reactants in the database"""
		self.quote_compounds(quoter=quoter, compounds=self.intermediates)

	### PLOTTING

	def plot_tag_statistics(self, *args, **kwargs):
		"""Plot an overview of the number of compounds and poses for each tag"""
		if not self.num_tags:
			logger.error('No tagged compounds or poses')
			return
		from .plotting import plot_tag_statistics
		return plot_tag_statistics(self, *args, **kwargs)

	def plot_compound_property(self, prop, **kwargs): 
		"""Plot an arbitrary compound property across the whole dataset"""
		from .plotting import plot_compound_property
		return plot_compound_property(self, prop, **kwargs)

	def plot_pose_property(self, prop, **kwargs): 
		"""Plot an arbitrary pose property across the whole dataset"""
		from .plotting import plot_pose_property
		return plot_pose_property(self, prop, **kwargs)

	def plot_interaction_punchcard(self, poses=None, subtitle=None, opacity=1.0, **kwargs):
		"""Plot an interaction punchcard for a set of poses"""
		from .plotting import plot_interaction_punchcard
		return plot_interaction_punchcard(self, poses=poses, subtitle=subtitle, opacity=opacity, **kwargs)

	def plot_residue_interactions(self, poses, residue_number, **kwargs):
		"""Plot an interaction punchcard for a set of poses"""
		from .plotting import plot_residue_interactions
		return plot_residue_interactions(self, poses=poses, residue_number=residue_number, **kwargs)

	def plot_compound_availability(self, compounds=None, **kwargs):
		"""Plot a bar chart of compound availability by supplier/catalogue"""
		from .plotting import plot_compound_availability
		return plot_compound_availability(self, compounds=compounds, **kwargs)

	def plot_compound_price(self, min_amount, compounds=None, plot_lead_time=False, style='histogram', **kwargs):
		"""Plot a bar chart of minimum compound price for a given minimum amount"""
		from .plotting import plot_compound_price
		return plot_compound_price(self, min_amount=min_amount, compounds=compounds, style=style, **kwargs)

	def plot_reaction_funnel(self, **kwargs):
		from .plotting import plot_reaction_funnel
		return plot_reaction_funnel(self, **kwargs)
	
	### OTHER

	def summary(self):
		"""Print a summary of this HIPPO"""
		logger.header(self)
		logger.var('db_path', self.db_path)
		logger.var('#compounds', self.num_compounds)
		logger.var('#poses', self.num_poses)
		logger.var('#reactions', self.num_reactions)
		logger.var('#tags', self.num_tags)
		logger.var('tags', self.tags.unique)
		# logger.var('#products', len(self.products))

		

	### DUNDERS

	def __repr__(self) -> str:
		"""Returns a command line representation"""
		return f'HIPPO("{self.name}")'

class InvalidRowError(Exception):
	...
